from abc import ABC, abstractmethod
from collections import defaultdict

import h5py
import hdf5plugin
import numpy as np
from ophyd import Kind, Staged
from scipy.ndimage import gaussian_filter

from ophyd_devices.sim.sim_data import NoiseType
from ophyd_devices.sim.sim_signals import CustomSetableSignal
from ophyd_devices.utils.bec_device_base import BECDeviceBase


class DeviceProxy(BECDeviceBase, ABC):
    """DeviceProxy class inherits from BECDeviceBase.

    It is an abstract class that is meant to be used as a base class for all device proxies.
    The minimum requirement for a device proxy is to implement the _compute method.
    """

    def __init__(self, name, *args, device_manager=None, **kwargs):
        self.name = name
        self.device_manager = device_manager
        self.config = None
        self._lookup = defaultdict(dict)
        super().__init__(name, *args, device_manager=device_manager, **kwargs)
        self._signals = dict()

    @property
    def lookup(self):
        """lookup property"""
        return self._lookup

    @lookup.setter
    def lookup(self, update: dict) -> None:
        """lookup setter"""
        self._lookup.update(update)

    def _update_device_config(self, config: dict) -> None:
        """
        BEC will call this method on every object upon initializing devices to pass over the deviceConfig
        from the config file. It can be conveniently be used to hand over initial parameters to the device.

        Args:
            config (dict): Config dictionary.
        """
        self.config = config
        self._compile_lookup()

    def _compile_lookup(self):
        """Compile the lookup table for the device."""
        for device_name in self.config.keys():
            self._lookup[device_name] = {
                "method": self._compute,
                "signal_name": self.config[device_name]["signal_name"],
                "args": (device_name,),
                "kwargs": {},
            }

    @abstractmethod
    def _compute(self, device_name: str, *args, **kwargs) -> any:
        """
        The purpose of this method is to compute the readback value for the signal of the device
        that this proxy is attached to. This method is meant to be overriden by the user.
        P

        Args:
            device_name (str): Name of the device.

        Returns:
        """


class SlitProxy(DeviceProxy):
    """
    Simulation framework to immidate the behaviour of slits.

    This device is a proxy that is meant to overrides the behaviour of a SimCamera.
    You may use this to simulate the effect of slits on the camera image.

    Parameters can be configured via the deviceConfig field in the device_config.
    The example below shows the configuration for a pinhole simulation on an Eiger detector,
    where the pinhole is defined by the position of motors samx and samy. These devices must
    exist in your config.

    To update for instance the pixel_size directly, you can directly access the DeviceConfig via
    `dev.eiger.get_device_config()` or update it `dev.eiger.get_device_config({'eiger' : {'pixel_size': 0.1}})`

    slit_sim:
        readoutPriority: baseline
        deviceClass: SlitProxy
        deviceConfig:
            eiger:
                signal_name: image
                center_offset: [0, 0] # [x,y]
                covariance: [[1000, 500], [200, 1000]] # [[x,x],[y,y]]
                pixel_size: 0.01
                ref_motors: [samx, samy]
                slit_width: [1, 1]
                motor_dir: [0, 1] # x:0 , y:1, z:2 coordinates
        enabled: true
        readOnly: false
    """

    USER_ACCESS = ["enabled", "lookup", "help"]

    def __init__(self, name, *args, device_manager=None, **kwargs):
        self._gaussian_blur_sigma = 5
        super().__init__(name, *args, device_manager=device_manager, **kwargs)

    def help(self) -> None:
        """Print documentation for the SlitLookup device."""
        print(self.__doc__)

    def _compute(self, device_name: str, *args, **kwargs) -> np.ndarray:
        """
        Compute the lookup table for the simulated camera.
        It copies the sim_camera bevahiour and adds a mask to simulate the effect of a pinhole.

        Args:
            device_name (str): Name of the device.
            signal_name (str): Name of the signal.

        Returns:
            np.ndarray: Lookup table for the simulated camera.
        """
        device_obj = self.device_manager.devices.get(device_name).obj
        params = device_obj.sim.sim_params
        shape = device_obj.image_shape.get()
        params.update(
            {
                "noise": NoiseType.POISSON,
                "covariance": np.array(self.config[device_name]["covariance"]),
                "center_offset": np.array(self.config[device_name]["center_offset"]),
            }
        )
        amp = params.get("amplitude")
        cov = params.get("covariance")
        cen_off = params.get("center_offset")

        pos, offset, cov, amp = device_obj.sim._prepare_params_gauss(
            amp=amp, cov=cov, offset=cen_off, shape=shape
        )
        v = device_obj.sim._compute_multivariate_gaussian(pos=pos, cen_off=offset, cov=cov, amp=amp)
        device_pos = self.config[device_name]["pixel_size"] * pos
        valid_mask = self._create_mask(
            device_pos=device_pos,
            ref_motors=self.config[device_name]["ref_motors"],
            width=self.config[device_name]["slit_width"],
            direction=self.config[device_name]["motor_dir"],
        )
        valid_mask = self._blur_image(valid_mask, sigma=self._gaussian_blur_sigma)
        v *= valid_mask
        v = device_obj.sim._add_noise(
            v, noise=params["noise"], noise_multiplier=params["noise_multiplier"]
        )
        v = device_obj.sim._add_hot_pixel(
            v,
            coords=params["hot_pixel_coords"],
            hot_pixel_types=params["hot_pixel_types"],
            values=params["hot_pixel_values"],
        )
        return v

    def _blur_image(self, image: np.ndarray, sigma: float = 1) -> np.ndarray:
        """Blur the image with a gaussian filter.

        Args:
            image (np.ndarray): Image to be blurred.
            sigma (float): Sigma for the gaussian filter.

        Returns:
            np.ndarray: Blurred image.
        """
        return gaussian_filter(image, sigma=sigma)

    def _create_mask(
        self,
        device_pos: np.ndarray,
        ref_motors: list[str],
        width: list[float],
        direction: list[int],
    ):
        mask = np.ones_like(device_pos)
        for ii, motor_name in enumerate(ref_motors):
            motor_pos = self.device_manager.devices.get(motor_name).obj.read()[motor_name]["value"]
            edges = [motor_pos + width[ii] / 2, motor_pos - width[ii] / 2]
            mask[..., direction[ii]] = np.logical_and(
                device_pos[..., direction[ii]] > np.min(edges),
                device_pos[..., direction[ii]] < np.max(edges),
            )

        return np.prod(mask, axis=2)


class H5ImageReplayProxy(DeviceProxy):
    """This Proxy class can be used to replay images from an h5 file.

    If the number of requested images is larger than the number of available iamges, the images will be replayed from the beginning.

    h5_image_sim:
        readoutPriority: baseline
        deviceClass: H5ImageReplayProxy
        deviceConfig:
            eiger:
                signal_name: image
                file_source: /path/to/h5file.h5
                h5_entry: /entry/data
        enabled: true
        readOnly: false
    """

    USER_ACCESS = ["file_source", "h5_entry"]

    def __init__(self, name, *args, device_manager=None, **kwargs):
        self.h5_file = None
        self.h5_dataset = None
        self._number_of_images = None
        self._staged = Staged.no
        self._image = None
        self._index = 0
        super().__init__(name, *args, device_manager=device_manager, **kwargs)
        self.file_source = CustomSetableSignal(
            name="file_source", value="", parent=self, kind=Kind.normal
        )
        self.h5_entry = CustomSetableSignal(
            name="h5_entry", value="", parent=self, kind=Kind.normal
        )

    @property
    def component_names(self) -> list[str]:
        """Return the names of the components."""
        return ["file_source", "h5_entry"]

    def _update_device_config(self, config: dict) -> None:
        super()._update_device_config(config)
        if len(config.keys()) > 1:
            raise RuntimeError(
                f"The current implementation of device {self.name} can only replay data for a single device. The config has information about multiple devices {config.keys()}"
            )
        self._init_signals()

    def _init_signals(self):
        """Initialize the signals for the device."""
        if "file_source" in self.config[list(self.config.keys())[0]]:
            self.file_source.set(self.config[list(self.config.keys())[0]]["file_source"])
        if "h5_entry" in self.config[list(self.config.keys())[0]]:
            self.h5_entry.set(self.config[list(self.config.keys())[0]]["h5_entry"])

    def _open_h5_file(self) -> None:
        """Opens the HDF5 file found in the file_source signal and the HDF5 dataset specified by the h5_entry signal."""
        self.h5_file = h5py.File(self.file_source.get(), mode="r")
        self.h5_dataset = self.h5_file[self.h5_entry.get()]
        self._number_of_images = self.h5_dataset.shape[0]

    def _close_h5_file(self) -> None:
        """Close the HDF5 file."""
        self.h5_file.close()

    def stop(self) -> None:
        """Stop the device."""
        if self.h5_file:
            self._close_h5_file()
        self.h5_file = None
        self.h5_dataset = None
        self._number_of_images = None
        self._index = 0

    def stage(self) -> list[object]:
        """Stage the device.
        This opens the HDF5 dataset, unstaging will close it.
        """

        if self._staged != Staged.no:
            return [self]
        try:
            self._open_h5_file()
        except Exception as exc:
            if self.h5_file:
                self.stop()
            raise FileNotFoundError(
                f"Could not open h5file {self.file_source.get()} or access data set {self.h5_dataset.get()} in file"
            ) from exc

        self._staged = Staged.yes
        return [self]

    def unstage(self) -> list[object]:
        """Unstage the device, also closes the HDF5 dataset"""
        if self.h5_file:
            self.stop()
        self._staged = Staged.no
        return [self]

    def _load_image(self):
        """Try loading the image from the h5 dataset, and set it to self._image."""
        if self.h5_file:
            slice_nr = self._index % self._number_of_images
            self._index = self._index + 1
            self._image = self.h5_dataset[slice_nr, ...]
            return
        try:
            self.stage()
            slice_nr = self._index % self._number_of_images
            self._index = self._index + 1
            self._image = self.h5_dataset[slice_nr, ...]
            self.unstage()
        except Exception as exc:
            raise FileNotFoundError(
                f"Could not open h5file {self.file_source.get()} or access data set {self.h5_dataset.get()} in file"
            ) from exc

    def _compute(self, device_name: str, *args, **kwargs) -> np.ndarray:
        """Compute the image.

        Returns:
            np.ndarray: Image.
        """
        self._load_image()
        return self._image


if __name__ == "__main__":
    # Example usage
    tmp = H5ImageReplayProxy(name="tmp", device_manager=None)
    config = {
        "eiger": {
            "signal_name": "image",
            "file_source": "/Users/appel_c/switchdrive/Sharefolder/AgBH_2D_gridscan/projection_000006_data_000001.h5",
            "h5_entry": "/entry/data/data",
        }
    }
    tmp._update_device_config(config)
    tmp.stage()
    print(tmp)



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 13 18:06:15 2021

@author: mohacsi_i

IMPORTANT: Virtual monochromator axes should be implemented already in EPICS!!!
"""

import time
from math import asin, atan, isclose, sin, sqrt, tan

import numpy as np
from ophyd import (
    Component,
    Device,
    EpicsMotor,
    EpicsSignal,
    EpicsSignalRO,
    Kind,
    PseudoPositioner,
    PseudoSingle,
    PVPositioner,
    Signal,
)
from ophyd.pseudopos import pseudo_position_argument, real_position_argument


class PmMonoBender(PseudoPositioner):
    """Monochromator bender

    Small wrapper to combine the four monochromator bender motors.
    """

    # Real axes
    ai = Component(EpicsMotor, "TRYA", name="ai")
    bo = Component(EpicsMotor, "TRYB", name="bo")
    co = Component(EpicsMotor, "TRYC", name="co")
    di = Component(EpicsMotor, "TRYD", name="di")

    # Virtual axis
    bend = Component(PseudoSingle, name="bend")

    _real = ["ai", "bo", "co", "di"]

    @pseudo_position_argument
    def forward(self, pseudo_pos):
        delta = pseudo_pos.bend - 0.25 * (
            self.ai.position + self.bo.position + self.co.position + self.di.position
        )
        return self.RealPosition(
            ai=self.ai.position + delta,
            bo=self.bo.position + delta,
            co=self.co.position + delta,
            di=self.di.position + delta,
        )

    @real_position_argument
    def inverse(self, real_pos):
        return self.PseudoPosition(
            bend=0.25 * (real_pos.ai + real_pos.bo + real_pos.co + real_pos.di)
        )


def r2d(radians):
    return radians * 180 / 3.141592


def d2r(degrees):
    return degrees * 3.141592 / 180.0


class PmDetectorRotation(PseudoPositioner):
    """Detector rotation pseudo motor

    Small wrapper to convert detector pusher position to rotation angle.
    """

    _tables_dt_push_dist_mm = 890
    # Real axes
    dtpush = Component(EpicsMotor, "", name="dtpush")

    # Virtual axis
    dtth = Component(PseudoSingle, name="dtth")

    _real = ["dtpush"]

    @pseudo_position_argument
    def forward(self, pseudo_pos):
        return self.RealPosition(
            dtpush=d2r(tan(-3.14 / 180 * pseudo_pos.dtth)) * self._tables_dt_push_dist_mm
        )

    @real_position_argument
    def inverse(self, real_pos):
        return self.PseudoPosition(dtth=r2d(-atan(real_pos.dtpush / self._tables_dt_push_dist_mm)))


class GirderMotorX1(PVPositioner):
    """Girder X translation pseudo motor"""

    setpoint = Component(EpicsSignal, ":X_SET", name="sp")
    readback = Component(EpicsSignalRO, ":X1", name="rbv")
    done = Component(EpicsSignal, ":M-DMOV", name="dmov")


class GirderMotorY1(PVPositioner):
    """Girder Y translation pseudo motor"""

    setpoint = Component(EpicsSignal, ":Y_SET", name="sp")
    readback = Component(EpicsSignalRO, ":Y1", name="rbv")
    done = Component(EpicsSignal, ":M-DMOV", name="dmov")


class GirderMotorYAW(PVPositioner):
    """Girder YAW pseudo motor"""

    setpoint = Component(EpicsSignal, ":YAW_SET", name="sp")
    readback = Component(EpicsSignalRO, ":YAW1", name="rbv")
    done = Component(EpicsSignal, ":M-DMOV", name="dmov")


class GirderMotorROLL(PVPositioner):
    """Girder ROLL pseudo motor"""

    setpoint = Component(EpicsSignal, ":ROLL_SET", name="sp")
    readback = Component(EpicsSignalRO, ":ROLL1", name="rbv")
    done = Component(EpicsSignal, ":M-DMOV", name="dmov")


class GirderMotorPITCH(PVPositioner):
    """Girder YAW pseudo motor"""

    setpoint = Component(EpicsSignal, ":PITCH_SET", name="sp")
    readback = Component(EpicsSignalRO, ":PITCH1", name="rbv")
    done = Component(EpicsSignal, ":M-DMOV", name="dmov")


class VirtualEpicsSignalRO(EpicsSignalRO):
    """This is a test class to create derives signals from one or
    multiple original signals...
    """

    def calc(self, val):
        return val

    def get(self, *args, **kwargs):
        raw = super().get(*args, **kwargs)
        return self.calc(raw)


class MonoTheta1(VirtualEpicsSignalRO):
    """Converts the pusher motor position to theta angle"""

    _mono_a0_enc_scale1 = -1.0
    _mono_a1_lever_length1 = 206.706
    _mono_a2_pusher_offs1 = 6.85858
    _mono_a3_enc_offs1 = -16.9731

    def calc(self, val):
        asin_arg = (val - self._mono_a2_pusher_offs1) / self._mono_a1_lever_length1
        theta1 = (
            self._mono_a0_enc_scale1 * asin(asin_arg) / 3.141592 * 180.0 + self._mono_a3_enc_offs1
        )
        return theta1


class MonoTheta2(VirtualEpicsSignalRO):
    """Converts the pusher motor position to theta angle"""

    _mono_a3_enc_offs2 = -19.7072
    _mono_a2_pusher_offs2 = 5.93905
    _mono_a1_lever_length2 = 206.572
    _mono_a0_enc_scale2 = -1.0

    def calc(self, val):
        asin_arg = (val - self._mono_a2_pusher_offs2) / self._mono_a1_lever_length2
        theta2 = (
            self._mono_a0_enc_scale2 * asin(asin_arg) / 3.141592 * 180.0 + self._mono_a3_enc_offs2
        )
        return theta2


MONO_THETA2_OFFSETS_FILENAME = (
    "/sls/X12SA/data/gac-x12saop/spec/macros/spec_data/mono_th2_offsets.txt"
)


class EnergyKev(VirtualEpicsSignalRO):
    """Converts the pusher motor position to energy in keV"""

    _mono_add_offs = True
    _mono_a3_enc_offs2 = -19.7072
    _mono_a2_pusher_offs2 = 5.93905
    _mono_a1_lever_length2 = 206.572
    _mono_a0_enc_scale2 = -1.0
    _mono_hce = 12.39852066
    _mono_2d2 = 2 * 5.43102 / sqrt(3)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._th2_offsets = np.loadtxt(MONO_THETA2_OFFSETS_FILENAME)

    def _mono_get_th2_offs(self, energy_keV):
        if self._th2_offsets is None:
            return 0.0

        max_offs = np.max(self._th2_offsets[:, 1])

        if max_offs > 0.2:
            raise ValueError(
                f"\nThe empirical moth2 corrections are as high as {max_offs} deg\nThis is unreasonable and the corrections will not be used.\n\n***PLEASE INFORM BEAMLINE SCIENTISTS***\n"
            )

        offs = np.interp(energy_keV, self._th2_offsets[:, 0], self._th2_offsets[:, 1])
        # print(offs)
        return offs

    def calc(self, val):
        _mono_sintheta2_to_Ekev = -self._mono_hce / self._mono_2d2
        asin_arg = (val - self._mono_a2_pusher_offs2) / self._mono_a1_lever_length2
        theta2_deg = (
            self._mono_a0_enc_scale2 * asin(asin_arg) / 3.141592 * 180.0 + self._mono_a3_enc_offs2
        )
        E_keV = _mono_sintheta2_to_Ekev / sin(theta2_deg / 180.0 * 3.141592)

        if self._mono_add_offs:
            theta2_deg -= self._mono_get_th2_offs(E_keV)
            E_keV = _mono_sintheta2_to_Ekev / sin(theta2_deg / 180.0 * 3.141592)
        return E_keV


class CurrentSum(Signal):
    """Adds up four current signals from the parent"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parent.ch1.subscribe(self._emit_value)

    def _emit_value(self, **kwargs):
        timestamp = kwargs.pop("timestamp", time.time())
        self.wait_for_connection()
        self._run_subs(sub_type="value", timestamp=timestamp, obj=self)

    def get(self, *args, **kwargs):
        # self.parent._cnt.set(1).wait()
        self._metadata["timestamp"] = time.time()
        total = (
            self.parent.ch1.get()
            + self.parent.ch2.get()
            + self.parent.ch3.get()
            + self.parent.ch4.get()
        )
        return total


class Bpm4i(Device):
    SUB_VALUE = "value"
    _default_sub = SUB_VALUE
    _cont = Component(EpicsSignal, "CONT", put_complete=True, kind=Kind.omitted)
    _cnt = Component(EpicsSignal, "CNT", put_complete=True, kind=Kind.omitted)
    ch1 = Component(EpicsSignalRO, "S2", auto_monitor=True, kind=Kind.omitted, name="ch1")
    ch2 = Component(EpicsSignalRO, "S3", auto_monitor=True, kind=Kind.omitted, name="ch2")
    ch3 = Component(EpicsSignalRO, "S4", auto_monitor=True, kind=Kind.omitted, name="ch3")
    ch4 = Component(EpicsSignalRO, "S5", auto_monitor=True, kind=Kind.omitted, name="ch4")
    sum = Component(CurrentSum, kind=Kind.hinted, name="sum")

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        **kwargs,
    ):
        super().__init__(
            prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        self.sum.name = self.name
        # Ensure the scaler counts automatically
        self._cont.wait_for_connection()
        self._cont.set(1).wait()
        self.ch1.subscribe(self._emit_value)

    def _emit_value(self, **kwargs):
        timestamp = kwargs.pop("timestamp", time.time())
        self.wait_for_connection()
        self._run_subs(sub_type=self.SUB_VALUE, timestamp=timestamp, obj=self)


if __name__ == "__main__":
    dut = Bpm4i("X12SA-OP1-SCALER.", name="bpm4")
    dut.wait_for_connection()
    print(dut.read())
    print(dut.describe())



from math import inf
from unittest import mock

import pytest

from bec_lib import MessageEndpoints, messages
from bec_lib.dap_plugin_objects import DAPPluginObject, LmfitService1D
from bec_lib.dap_plugins import DAPPlugins
from bec_lib.device import DeviceBase
from bec_lib.scan_items import ScanItem
from bec_lib.scan_report import ScanReport


@pytest.fixture
def dap_plugin_message():
    msg = messages.AvailableResourceMessage(
        **{
            "resource": {
                "GaussianModel": {
                    "class": "LmfitService1D",
                    "user_friendly_name": "GaussianModel",
                    "class_doc": "A model based on a Gaussian or normal distribution lineshape.\n\n    The model has three Parameters: `amplitude`, `center`, and `sigma`.\n    In addition, parameters `fwhm` and `height` are included as\n    constraints to report full width at half maximum and maximum peak\n    height, respectively.\n\n    .. math::\n\n        f(x; A, \\mu, \\sigma) = \\frac{A}{\\sigma\\sqrt{2\\pi}} e^{[{-{(x-\\mu)^2}/{{2\\sigma}^2}}]}\n\n    where the parameter `amplitude` corresponds to :math:`A`, `center` to\n    :math:`\\mu`, and `sigma` to :math:`\\sigma`. The full width at half\n    maximum is :math:`2\\sigma\\sqrt{2\\ln{2}}`, approximately\n    :math:`2.3548\\sigma`.\n\n    For more information, see: https://en.wikipedia.org/wiki/Normal_distribution\n\n    ",
                    "run_doc": "A model based on a Gaussian or normal distribution lineshape.\n\n    The model has three Parameters: `amplitude`, `center`, and `sigma`.\n    In addition, parameters `fwhm` and `height` are included as\n    constraints to report full width at half maximum and maximum peak\n    height, respectively.\n\n    .. math::\n\n        f(x; A, \\mu, \\sigma) = \\frac{A}{\\sigma\\sqrt{2\\pi}} e^{[{-{(x-\\mu)^2}/{{2\\sigma}^2}}]}\n\n    where the parameter `amplitude` corresponds to :math:`A`, `center` to\n    :math:`\\mu`, and `sigma` to :math:`\\sigma`. The full width at half\n    maximum is :math:`2\\sigma\\sqrt{2\\ln{2}}`, approximately\n    :math:`2.3548\\sigma`.\n\n    For more information, see: https://en.wikipedia.org/wiki/Normal_distribution\n\n    \n        Args:\n            scan_item (ScanItem): Scan item or scan ID\n            device_x (DeviceBase | str): Device name for x\n            signal_x (DeviceBase | str): Signal name for x\n            device_y (DeviceBase | str): Device name for y\n            signal_y (DeviceBase | str): Signal name for y\n            parameters (dict): Fit parameters\n        ",
                    "run_name": "fit",
                    "signature": [
                        {
                            "name": "args",
                            "kind": "VAR_POSITIONAL",
                            "default": "_empty",
                            "annotation": "_empty",
                        },
                        {
                            "name": "scan_item",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "ScanItem | str",
                        },
                        {
                            "name": "device_x",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "DeviceBase | str",
                        },
                        {
                            "name": "signal_x",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "DeviceBase | str",
                        },
                        {
                            "name": "device_y",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "DeviceBase | str",
                        },
                        {
                            "name": "signal_y",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "DeviceBase | str",
                        },
                        {
                            "name": "parameters",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "dict",
                        },
                        {
                            "name": "kwargs",
                            "kind": "VAR_KEYWORD",
                            "default": "_empty",
                            "annotation": "_empty",
                        },
                    ],
                    "auto_fit_supported": True,
                    "params": {
                        "amplitude": {
                            "name": "amplitude",
                            "value": 1.0,
                            "vary": True,
                            "min": -inf,
                            "max": inf,
                            "expr": None,
                            "brute_step": None,
                            "user_data": None,
                        },
                        "center": {
                            "name": "center",
                            "value": 0.0,
                            "vary": True,
                            "min": -inf,
                            "max": inf,
                            "expr": None,
                            "brute_step": None,
                            "user_data": None,
                        },
                        "sigma": {
                            "name": "sigma",
                            "value": 1.0,
                            "vary": True,
                            "min": 0,
                            "max": inf,
                            "expr": None,
                            "brute_step": None,
                            "user_data": None,
                        },
                        "fwhm": {
                            "name": "fwhm",
                            "value": 2.35482,
                            "vary": False,
                            "min": -inf,
                            "max": inf,
                            "expr": "2.3548200*sigma",
                            "brute_step": None,
                            "user_data": None,
                        },
                        "height": {
                            "name": "height",
                            "value": 0.3989423,
                            "vary": False,
                            "min": -inf,
                            "max": inf,
                            "expr": "0.3989423*amplitude/max(1e-15, sigma)",
                            "brute_step": None,
                            "user_data": None,
                        },
                    },
                    "class_args": [],
                    "class_kwargs": {"model": "GaussianModel"},
                },
                "StepModel": {
                    "class": "LmfitService1D",
                    "user_friendly_name": "StepModel",
                    "class_doc": "A model based on a Step function.\n\n    The model has three Parameters: `amplitude` (:math:`A`), `center`\n    (:math:`\\mu`), and `sigma` (:math:`\\sigma`).\n\n    There are four choices for `form`:\n\n    - `'linear'` (default)\n    - `'atan'` or `'arctan'` for an arc-tangent function\n    - `'erf'` for an error function\n    - `'logistic'` for a logistic function (for more information, see:\n      https://en.wikipedia.org/wiki/Logistic_function)\n\n    The step function starts with a value 0 and ends with a value of\n    :math:`A` rising to :math:`A/2` at :math:`\\mu`, with :math:`\\sigma`\n    setting the characteristic width. The functional forms are defined as:\n\n    .. math::\n        :nowrap:\n\n        \\begin{eqnarray*}\n        & f(x; A, \\mu, \\sigma, {\\mathrm{form={}'linear{}'}})  & = A \\min{[1, \\max{(0, \\alpha + 1/2)}]} \\\\\n        & f(x; A, \\mu, \\sigma, {\\mathrm{form={}'arctan{}'}})  & = A [1/2 + \\arctan{(\\alpha)}/{\\pi}] \\\\\n        & f(x; A, \\mu, \\sigma, {\\mathrm{form={}'erf{}'}})     & = A [1 + {\\operatorname{erf}}(\\alpha)]/2 \\\\\n        & f(x; A, \\mu, \\sigma, {\\mathrm{form={}'logistic{}'}})& = A \\left[1 - \\frac{1}{1 + e^{\\alpha}} \\right]\n        \\end{eqnarray*}\n\n    where :math:`\\alpha = (x - \\mu)/{\\sigma}`.\n\n    ",
                    "run_doc": "A model based on a Step function.\n\n    The model has three Parameters: `amplitude` (:math:`A`), `center`\n    (:math:`\\mu`), and `sigma` (:math:`\\sigma`).\n\n    There are four choices for `form`:\n\n    - `'linear'` (default)\n    - `'atan'` or `'arctan'` for an arc-tangent function\n    - `'erf'` for an error function\n    - `'logistic'` for a logistic function (for more information, see:\n      https://en.wikipedia.org/wiki/Logistic_function)\n\n    The step function starts with a value 0 and ends with a value of\n    :math:`A` rising to :math:`A/2` at :math:`\\mu`, with :math:`\\sigma`\n    setting the characteristic width. The functional forms are defined as:\n\n    .. math::\n        :nowrap:\n\n        \\begin{eqnarray*}\n        & f(x; A, \\mu, \\sigma, {\\mathrm{form={}'linear{}'}})  & = A \\min{[1, \\max{(0, \\alpha + 1/2)}]} \\\\\n        & f(x; A, \\mu, \\sigma, {\\mathrm{form={}'arctan{}'}})  & = A [1/2 + \\arctan{(\\alpha)}/{\\pi}] \\\\\n        & f(x; A, \\mu, \\sigma, {\\mathrm{form={}'erf{}'}})     & = A [1 + {\\operatorname{erf}}(\\alpha)]/2 \\\\\n        & f(x; A, \\mu, \\sigma, {\\mathrm{form={}'logistic{}'}})& = A \\left[1 - \\frac{1}{1 + e^{\\alpha}} \\right]\n        \\end{eqnarray*}\n\n    where :math:`\\alpha = (x - \\mu)/{\\sigma}`.\n\n    \n        Args:\n            scan_item (ScanItem): Scan item or scan ID\n            device_x (DeviceBase | str): Device name for x\n            signal_x (DeviceBase | str): Signal name for x\n            device_y (DeviceBase | str): Device name for y\n            signal_y (DeviceBase | str): Signal name for y\n            parameters (dict): Fit parameters\n        ",
                    "run_name": "fit",
                    "signature": [
                        {
                            "name": "args",
                            "kind": "VAR_POSITIONAL",
                            "default": "_empty",
                            "annotation": "_empty",
                        },
                        {
                            "name": "scan_item",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "ScanItem | str",
                        },
                        {
                            "name": "device_x",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "DeviceBase | str",
                        },
                        {
                            "name": "signal_x",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "DeviceBase | str",
                        },
                        {
                            "name": "device_y",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "DeviceBase | str",
                        },
                        {
                            "name": "signal_y",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "DeviceBase | str",
                        },
                        {
                            "name": "parameters",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "dict",
                        },
                        {
                            "name": "kwargs",
                            "kind": "VAR_KEYWORD",
                            "default": "_empty",
                            "annotation": "_empty",
                        },
                    ],
                    "auto_fit_supported": True,
                    "params": {
                        "amplitude": {
                            "name": "amplitude",
                            "value": 1.0,
                            "vary": True,
                            "min": -inf,
                            "max": inf,
                            "expr": None,
                            "brute_step": None,
                            "user_data": None,
                        },
                        "center": {
                            "name": "center",
                            "value": 0.0,
                            "vary": True,
                            "min": -inf,
                            "max": inf,
                            "expr": None,
                            "brute_step": None,
                            "user_data": None,
                        },
                        "sigma": {
                            "name": "sigma",
                            "value": 1.0,
                            "vary": True,
                            "min": -inf,
                            "max": inf,
                            "expr": None,
                            "brute_step": None,
                            "user_data": None,
                        },
                    },
                    "class_args": [],
                    "class_kwargs": {"model": "StepModel"},
                },
                "Ptychography": {
                    "class": "PtychographyDAP",
                    "user_friendly_name": "ptycho",
                    "class_doc": "A model based on a Gaussian or normal distribution lineshape.\n\n    The model has three Parameters: `amplitude`, `center`, and `sigma`.\n    In addition, parameters `fwhm` and `height` are included as\n    constraints to report full width at half maximum and maximum peak\n    height, respectively.\n\n    .. math::\n\n        f(x; A, \\mu, \\sigma) = \\frac{A}{\\sigma\\sqrt{2\\pi}} e^{[{-{(x-\\mu)^2}/{{2\\sigma}^2}}]}\n\n    where the parameter `amplitude` corresponds to :math:`A`, `center` to\n    :math:`\\mu`, and `sigma` to :math:`\\sigma`. The full width at half\n    maximum is :math:`2\\sigma\\sqrt{2\\ln{2}}`, approximately\n    :math:`2.3548\\sigma`.\n\n    For more information, see: https://en.wikipedia.org/wiki/Normal_distribution\n\n    ",
                    "run_doc": "A model based on a Gaussian or normal distribution lineshape.\n\n    The model has three Parameters: `amplitude`, `center`, and `sigma`.\n    In addition, parameters `fwhm` and `height` are included as\n    constraints to report full width at half maximum and maximum peak\n    height, respectively.\n\n    .. math::\n\n        f(x; A, \\mu, \\sigma) = \\frac{A}{\\sigma\\sqrt{2\\pi}} e^{[{-{(x-\\mu)^2}/{{2\\sigma}^2}}]}\n\n    where the parameter `amplitude` corresponds to :math:`A`, `center` to\n    :math:`\\mu`, and `sigma` to :math:`\\sigma`. The full width at half\n    maximum is :math:`2\\sigma\\sqrt{2\\ln{2}}`, approximately\n    :math:`2.3548\\sigma`.\n\n    For more information, see: https://en.wikipedia.org/wiki/Normal_distribution\n\n    \n        Args:\n            scan_item (ScanItem): Scan item or scan ID\n            device_x (DeviceBase | str): Device name for x\n            signal_x (DeviceBase | str): Signal name for x\n            device_y (DeviceBase | str): Device name for y\n            signal_y (DeviceBase | str): Signal name for y\n            parameters (dict): Fit parameters\n        ",
                    "run_name": "fit",
                    "signature": [
                        {
                            "name": "args",
                            "kind": "VAR_POSITIONAL",
                            "default": "_empty",
                            "annotation": "_empty",
                        },
                        {
                            "name": "scan_item",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "ScanItem | str",
                        },
                        {
                            "name": "device_x",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "DeviceBase | str",
                        },
                        {
                            "name": "signal_x",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "DeviceBase | str",
                        },
                        {
                            "name": "device_y",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "DeviceBase | str",
                        },
                        {
                            "name": "signal_y",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "DeviceBase | str",
                        },
                        {
                            "name": "parameters",
                            "kind": "KEYWORD_ONLY",
                            "default": None,
                            "annotation": "dict",
                        },
                        {
                            "name": "kwargs",
                            "kind": "VAR_KEYWORD",
                            "default": "_empty",
                            "annotation": "_empty",
                        },
                    ],
                    "auto_fit_supported": True,
                    "params": {
                        "amplitude": {
                            "name": "amplitude",
                            "value": 1.0,
                            "vary": True,
                            "min": -inf,
                            "max": inf,
                            "expr": None,
                            "brute_step": None,
                            "user_data": None,
                        },
                        "center": {
                            "name": "center",
                            "value": 0.0,
                            "vary": True,
                            "min": -inf,
                            "max": inf,
                            "expr": None,
                            "brute_step": None,
                            "user_data": None,
                        },
                        "sigma": {
                            "name": "sigma",
                            "value": 1.0,
                            "vary": True,
                            "min": 0,
                            "max": inf,
                            "expr": None,
                            "brute_step": None,
                            "user_data": None,
                        },
                        "fwhm": {
                            "name": "fwhm",
                            "value": 2.35482,
                            "vary": False,
                            "min": -inf,
                            "max": inf,
                            "expr": "2.3548200*sigma",
                            "brute_step": None,
                            "user_data": None,
                        },
                        "height": {
                            "name": "height",
                            "value": 0.3989423,
                            "vary": False,
                            "min": -inf,
                            "max": inf,
                            "expr": "0.3989423*amplitude/max(1e-15, sigma)",
                            "brute_step": None,
                            "user_data": None,
                        },
                    },
                    "class_args": [],
                    "class_kwargs": {"model": "GaussianModel"},
                },
            }
        }
    )
    yield msg


@pytest.fixture
def dap(dap_plugin_message):
    dap_services = {
        "BECClient": messages.StatusMessage(name="BECClient", status=1, info={}),
        "DAPServer/LmfitService1D": messages.StatusMessage(
            name="LmfitService1D", status=1, info={}
        ),
        "DAPServer/PtychographyDAP": messages.StatusMessage(
            name="PtychographyDAP", status=1, info={}
        ),
    }
    client = mock.MagicMock()
    client.service_status = dap_services
    client.connector.get.return_value = dap_plugin_message
    dap_plugins = DAPPlugins(client)
    yield dap_plugins


def test_dap_plugins_construction(dap):
    assert hasattr(dap, "GaussianModel")
    assert hasattr(dap, "StepModel")
    assert hasattr(dap, "ptycho")
    # pylint: disable=no-member
    assert isinstance(dap.GaussianModel, LmfitService1D)
    assert isinstance(dap.StepModel, DAPPluginObject)
    assert isinstance(dap.ptycho, DAPPluginObject)


def test_dap_plugin_fit(dap):
    with mock.patch.object(dap.GaussianModel, "_wait_for_dap_response") as mock_wait:
        dap.GaussianModel.fit()
        dap._parent.connector.set_and_publish.assert_called_once()
        mock_wait.assert_called_once()


class ScanReportMock(ScanReport):
    def __init__(self, scan_id: str) -> None:
        super().__init__()
        self.request = mock.MagicMock()
        self.request.scan.scan_id = scan_id


@pytest.mark.parametrize(
    "input",
    [
        "scan_id",
        ScanItem(
            mock.MagicMock(), queue_id="queue_id", scan_id="scan_id", scan_number=1, status="closed"
        ),
        ScanReportMock("scan_id"),
    ],
)
def test_dap_plugin_fit_input(dap, input):
    with mock.patch.object(dap.GaussianModel, "_wait_for_dap_response") as mock_wait:
        dap.GaussianModel.fit(input)
        request_id = dap._parent.connector.set_and_publish.call_args[0][1].metadata["RID"]
        dap._parent.connector.set_and_publish.assert_called_once_with(
            MessageEndpoints.dap_request(),
            messages.DAPRequestMessage(
                dap_cls="LmfitService1D",
                dap_type="on_demand",
                config={
                    "args": ["scan_id"],
                    "kwargs": {},
                    "class_args": [],
                    "class_kwargs": {"model": "GaussianModel"},
                },
                metadata={"RID": request_id},
            ),
        )
        mock_wait.assert_called_once()


def test_dap_auto_run(dap):
    with mock.patch.object(dap.GaussianModel, "_update_dap_config") as mock_update_dap_config:
        dap.GaussianModel.auto_run == False
        dap.GaussianModel.auto_run = True
        mock_update_dap_config.assert_called_once()
        dap.GaussianModel.auto_run = True


def test_dap_wait_for_dap_response_waits_for_RID(dap):
    dap._parent.connector.get.return_value = messages.DAPResponseMessage(
        success=True, data=({}, None), metadata={"RID": "wrong_ID"}
    )
    with pytest.raises(TimeoutError):
        dap.GaussianModel._wait_for_dap_response(request_id="1234", timeout=0.1)


def test_dap_wait_for_dap_respnse_returns(dap):
    dap._parent.connector.get.return_value = messages.DAPResponseMessage(
        success=True, data=({}, None), metadata={"RID": "1234"}
    )
    val = dap.GaussianModel._wait_for_dap_response(request_id="1234", timeout=0.1)
    assert val == messages.DAPResponseMessage(
        success=True, data=({}, None), metadata={"RID": "1234"}
    )


def test_dap_select(dap):
    with mock.patch.object(dap.GaussianModel, "_update_dap_config") as mock_update_dap_config:
        obj = DeviceBase(name="samx", info={"device_info": {"hints": {"fields": ["samx"]}}})
        dap._parent.device_manager.devices.get.return_value = obj
        dap.GaussianModel.select("samx")
        mock_update_dap_config.assert_called_once()
        assert dap.GaussianModel._plugin_config["selected_device"] == ["samx", "samx"]


def test_dap_select_raises_with_too_many_hints(dap):
    with pytest.raises(AttributeError):
        with mock.patch.object(dap.GaussianModel, "_update_dap_config") as mock_update_dap_config:
            obj = DeviceBase(
                name="samx", info={"device_info": {"hints": {"fields": ["samx", "samx2"]}}}
            )
            dap._parent.device_manager.devices.get.return_value = obj
            dap.GaussianModel.select("samx")
            mock_update_dap_config.assert_called_once()
            assert dap.GaussianModel._plugin_config["selected_device"] == ["samx", "samx"]


def test_dap_select_raises_on_device_without_hints(dap):
    with pytest.raises(AttributeError):
        dap._parent.device_manager.devices.get.return_value = DeviceBase(name="samx", info={})
        dap.GaussianModel.select("samx")


def test_dap_select_raises_on_wrong_device(dap):
    dap._parent.device_manager.devices.get.return_value = None
    with pytest.raises(AttributeError):
        dap.GaussianModel.select("samx")


def test_dap_get_data(dap):
    dap._parent.connector.get_last.return_value = messages.ProcessedDataMessage(
        data=[{"x": [1, 2, 3], "y": [4, 5, 6]}, {"fit_parameters": {"amplitude": 1}}]
    )
    data = dap.GaussianModel.get_data()
    dap._parent.connector.get_last.assert_called_once_with(
        MessageEndpoints.processed_data("GaussianModel")
    )

    assert data.data == {"x": [1, 2, 3], "y": [4, 5, 6]}
    assert data.amplitude == 1


def test_dap_update_dap_config_not_called_without_device(dap):
    dap.GaussianModel._update_dap_config(request_id="1234")
    dap._parent.connector.set_and_publish.assert_not_called()


def test_dap_update_dap_config(dap):
    dap.GaussianModel._plugin_config["selected_device"] = ["samx", "samx"]
    dap.GaussianModel._update_dap_config(request_id="1234")
    dap._parent.connector.set_and_publish.assert_called_with(
        MessageEndpoints.dap_request(),
        messages.DAPRequestMessage(
            dap_cls="LmfitService1D",
            dap_type="continuous",
            config=dap.GaussianModel._plugin_config,
            metadata={"RID": "1234"},
        ),
    )



"""
This module contains the numpy encoder and decoder functions for serializing and deserializing numpy objects.
Modified from https://github.com/lebedov/msgpack-numpy
"""

import pickle
import sys

import numpy as np


def ndarray_to_bytes(obj):
    if obj.dtype == "O":
        return obj.dumps()
    if sys.platform == "darwin":
        return obj.tobytes()
    return obj.data if obj.flags["C_CONTIGUOUS"] else obj.tobytes()


def tostr(x):
    if isinstance(x, bytes):
        return x.decode()
    return str(x)


def numpy_encode(obj, chain=None):
    """
    Data encoder for serializing numpy data types.
    """

    if isinstance(obj, np.ndarray):
        # If the dtype is structured, store the interface description;
        # otherwise, store the corresponding array protocol type string:
        if obj.dtype.kind in ("V", "O"):
            kind = bytes(obj.dtype.kind, "ascii")
            descr = obj.dtype.descr
        else:
            kind = b""
            descr = obj.dtype.str

        return {
            b"nd": True,
            b"type": descr,
            b"kind": kind,
            b"shape": obj.shape,
            b"data": ndarray_to_bytes(obj),
        }
    elif isinstance(obj, (np.bool_, np.number)):
        return {b"nd": False, b"type": obj.dtype.str, b"data": obj.data}
    elif isinstance(obj, complex):
        return {b"complex": True, b"data": repr(obj)}
    else:
        return obj if chain is None else chain(obj)


def numpy_decode(obj, chain=None):
    """
    Decoder for deserializing numpy data types.
    """

    try:
        if b"nd" in obj:
            if obj[b"nd"] is True:
                # Check if b'kind' is in obj to enable decoding of data
                # serialized with older versions (#20) or data
                # that had dtype == 'O' (#46):
                if b"kind" in obj and obj[b"kind"] == b"V":
                    descr = [
                        tuple(tostr(t) if type(t) is bytes else t for t in d) for d in obj[b"type"]
                    ]
                elif b"kind" in obj and obj[b"kind"] == b"O":
                    return pickle.loads(obj[b"data"])
                else:
                    descr = obj[b"type"]
                return np.ndarray(
                    buffer=obj[b"data"], dtype=_unpack_dtype(descr), shape=obj[b"shape"]
                )
            else:
                descr = obj[b"type"]
                return np.frombuffer(obj[b"data"], dtype=_unpack_dtype(descr))[0]
        elif b"complex" in obj:
            return complex(tostr(obj[b"data"]))
        else:
            return obj if chain is None else chain(obj)
    except KeyError:
        return obj if chain is None else chain(obj)


def numpy_encode_list(obj, chain=None):
    """
    Data encoder for serializing numpy data types.
    """

    if isinstance(obj, np.ndarray):
        # If the dtype is structured, store the interface description;
        # otherwise, store the corresponding array protocol type string:
        if obj.dtype.kind in ("V", "O"):
            kind = bytes(obj.dtype.kind, "ascii")
            descr = obj.dtype.descr
        else:
            kind = ""
            descr = obj.dtype.str

        return {"nd": True, "type": descr, "kind": kind, "shape": obj.shape, "data": obj.tolist()}
    if isinstance(obj, (np.bool_, np.number)):
        return {"nd": False, "type": obj.dtype.str, "data": obj.data}
    if isinstance(obj, complex):
        return {"complex": True, "data": repr(obj)}

    return obj if chain is None else chain(obj)


def numpy_decode_list(obj, chain=None):
    """
    Decoder for deserializing numpy data types.
    """

    try:
        if "nd" in obj:
            if obj["nd"] is True:
                # Check if 'kind' is in obj to enable decoding of data
                # serialized with older versions or data that had dtype == 'O':
                if "kind" in obj and obj["kind"] == "V":
                    descr = [
                        tuple(tostr(t) if type(t) is bytes else t for t in d) for d in obj["type"]
                    ]
                elif "kind" in obj and obj["kind"] == "O":
                    return np.array(obj["data"])
                else:
                    descr = obj["type"]
                return np.array(obj["data"], dtype=_unpack_dtype(descr)).reshape(obj["shape"])
            descr = obj["type"]
            return np.array(obj["data"], dtype=_unpack_dtype(descr))[0]
        if "complex" in obj:
            return complex(tostr(obj["data"]))
        return obj if chain is None else chain(obj)
    except KeyError:
        return obj if chain is None else chain(obj)


def _unpack_dtype(dtype):
    """
    Unpack dtype descr, recursively unpacking nested structured dtypes.
    """

    if isinstance(dtype, (list, tuple)):
        # Unpack structured dtypes of the form: (name, type, *shape)
        dtype = [
            (subdtype[0], _unpack_dtype(subdtype[1])) + tuple(subdtype[2:]) for subdtype in dtype
        ]
    return np.dtype(dtype)



import os

import bec_lib
from bec_lib.scibec_validator import SciBecValidator

dir_path = os.path.abspath(os.path.join(os.path.dirname(bec_lib.__file__), "./configs/"))
SCHEMA = os.path.join(dir_path, "openapi_schema.json")


def test_scibec_validator():
    validator = SciBecValidator(SCHEMA)
    # print(validator.device_schema)
    validator.validate_device_patch({"enabled": True})



import pytest

import bec_lib
from bec_lib import plugin_helper


@pytest.mark.parametrize(
    "class_spec, out_name",
    [("bec_lib.messages.BECMessage", "BECMessage"), ("bec_lib.BECStatus", "BECStatus")],
)
def test_get_plugin_class(class_spec, out_name):
    cls = plugin_helper.get_plugin_class(class_spec, [bec_lib])
    assert cls.__name__ == out_name


@pytest.mark.parametrize(
    "class_spec", ["bec_lib.nonexistent_module.NonexistentClass", "bec_lib.NonexistentClass"]
)
def test_get_plugin_class_module_not_found(class_spec):
    with pytest.raises((ModuleNotFoundError, AttributeError)):
        plugin_helper.get_plugin_class(class_spec, [bec_lib])



import ast
import enum
import threading
import time
from abc import ABC, abstractmethod
from typing import Any, Literal

import numpy as np

from bec_lib import DeviceManagerBase, MessageEndpoints, bec_logger

from .errors import LimitError, ScanAbortion
from .path_optimization import PathOptimizerMixin
from .scan_stubs import ScanStubs

logger = bec_logger.logger


class ScanArgType(str, enum.Enum):
    DEVICE = "device"
    FLOAT = "float"
    INT = "int"
    BOOL = "boolean"
    STR = "str"
    LIST = "list"
    DICT = "dict"


def unpack_scan_args(scan_args: dict[str, Any]) -> list:
    """unpack_scan_args unpacks the scan arguments and returns them as a tuple.

    Args:
        scan_args (dict[str, Any]): scan arguments

    Returns:
        list: list of arguments
    """
    args = []
    if not scan_args:
        return args
    if not isinstance(scan_args, dict):
        return scan_args
    for cmd_name, cmd_args in scan_args.items():
        args.append(cmd_name)
        args.extend(cmd_args)
    return args


def get_2D_raster_pos(axis, snaked=True):
    """get_2D_raster_post calculates and returns the positions for a 2D

    snaked==True:
        ->->->->-
        -<-<-<-<-
        ->->->->-
    snaked==False:
        ->->->->-
        ->->->->-
        ->->->->-

    Args:
        axis (list): list of positions for each axis
        snaked (bool, optional): If true, the positions will be calculcated for a snake scan. Defaults to True.

    Returns:
        array: calculated positions
    """

    x_grid, y_grid = np.meshgrid(axis[0], axis[1])
    if snaked:
        y_grid.T[::2] = np.fliplr(y_grid.T[::2])
    x_flat = x_grid.T.ravel()
    y_flat = y_grid.T.ravel()
    positions = np.vstack((x_flat, y_flat)).T
    return positions


# pylint: disable=too-many-arguments
def get_fermat_spiral_pos(
    m1_start, m1_stop, m2_start, m2_stop, step=1, spiral_type=0, center=False
):
    """get_fermat_spiral_pos calculates and returns the positions for a Fermat spiral scan.

    Args:
        m1_start (float): start position motor 1
        m1_stop (float): end position motor 1
        m2_start (float): start position motor 2
        m2_stop (float): end position motor 2
        step (float, optional): Step size. Defaults to 1.
        spiral_type (float, optional): Angular offset in radians that determines the shape of the spiral.
        A spiral with spiral_type=2 is the same as spiral_type=0. Defaults to 0.
        center (bool, optional): Add a center point. Defaults to False.

    Returns:
        array: calculated positions in the form [[m1, m2], ...]
    """
    positions = []
    phi = 2 * np.pi * ((1 + np.sqrt(5)) / 2.0) + spiral_type * np.pi

    start = int(not center)

    length_axis1 = abs(m1_stop - m1_start)
    length_axis2 = abs(m2_stop - m2_start)
    n_max = int(length_axis1 * length_axis2 * 3.2 / step / step)

    for ii in range(start, n_max):
        radius = step * 0.57 * np.sqrt(ii)
        if abs(radius * np.sin(ii * phi)) > length_axis1 / 2:
            continue
        if abs(radius * np.cos(ii * phi)) > length_axis2 / 2:
            continue
        positions.extend([(radius * np.sin(ii * phi), radius * np.cos(ii * phi))])
    return np.array(positions)


def get_round_roi_scan_positions(lx: float, ly: float, dr: float, nth: int, cenx=0, ceny=0):
    """
    get_round_roi_scan_positions calculates and returns the positions for a round scan in a rectangular region of interest.

    Args:
        lx (float): length in x
        ly (float): length in y
        dr (float): step size
        nth (int): number of angles in the inner ring
        cenx (int, optional): center in x. Defaults to 0.
        ceny (int, optional): center in y. Defaults to 0.

    Returns:
        array: calculated positions in the form [[x, y], ...]
    """
    positions = []
    nr = 1 + int(np.floor(max([lx, ly]) / dr))
    for ir in range(1, nr + 2):
        rr = ir * dr
        dth = 2 * np.pi / (nth * ir)
        pos = [
            (rr * np.cos(ith * dth) + cenx, rr * np.sin(ith * dth) + ceny)
            for ith in range(nth * ir)
            if np.abs(rr * np.cos(ith * dth)) < lx / 2 and np.abs(rr * np.sin(ith * dth)) < ly / 2
        ]
        positions.extend(pos)
    return np.array(positions)


def get_round_scan_positions(r_in: float, r_out: float, nr: int, nth: int, cenx=0, ceny=0):
    """
    get_round_scan_positions calculates and returns the positions for a round scan.

    Args:
        r_in (float): inner radius
        r_out (float): outer radius
        nr (int): number of radii
        nth (int): number of angles in the inner ring
        cenx (int, optional): center in x. Defaults to 0.
        ceny (int, optional): center in y. Defaults to 0.

    Returns:
        array: calculated positions in the form [[x, y], ...]

    """
    positions = []
    dr = (r_in - r_out) / nr
    for ir in range(1, nr + 2):
        rr = r_in + ir * dr
        dth = 2 * np.pi / (nth * ir)
        positions.extend(
            [
                (rr * np.sin(ith * dth) + cenx, rr * np.cos(ith * dth) + ceny)
                for ith in range(nth * ir)
            ]
        )
    return np.array(positions)


class RequestBase(ABC):
    """
    Base class for all scan requests.
    """

    scan_name = ""
    scan_report_hint = None
    arg_input = {"device": ScanArgType.DEVICE}
    arg_bundle_size = {"bundle": len(arg_input), "min": 1, "max": None}
    required_kwargs = []
    return_to_start_after_abort = False

    def __init__(
        self,
        *args,
        device_manager: DeviceManagerBase = None,
        monitored: list = None,
        parameter: dict = None,
        metadata: dict = None,
        **kwargs,
    ) -> None:
        super().__init__()
        self._shutdown_event = threading.Event()
        self.parameter = parameter if parameter is not None else {}
        self.caller_args = self.parameter.get("args", {})
        self.caller_kwargs = self.parameter.get("kwargs", {})
        self.metadata = metadata
        self.device_manager = device_manager
        self.connector = device_manager.connector
        self.DIID = 0
        self.scan_motors = []
        self.positions = []
        self._pre_scan_macros = []
        self._scan_report_devices = None
        self._get_scan_motors()
        self.readout_priority = {
            "monitored": monitored if monitored is not None else [],
            "baseline": [],
            "on_request": [],
            "async": [],
        }
        self.update_readout_priority()
        if metadata is None:
            self.metadata = {}
        self.stubs = ScanStubs(
            connector=self.device_manager.connector,
            device_msg_callback=self.device_msg_metadata,
            shutdown_event=self._shutdown_event,
        )

    @property
    def scan_report_devices(self):
        """devices to be included in the scan report"""
        if self._scan_report_devices is None:
            return self.readout_priority["monitored"]
        return self._scan_report_devices

    @scan_report_devices.setter
    def scan_report_devices(self, devices: list):
        self._scan_report_devices = devices

    def device_msg_metadata(self):
        default_metadata = {"readout_priority": "monitored", "DIID": self.DIID}
        metadata = {**default_metadata, **self.metadata}
        self.DIID += 1
        return metadata

    @staticmethod
    def _get_func_name_from_macro(macro: str):
        return ast.parse(macro).body[0].name

    def run_pre_scan_macros(self):
        """run pre scan macros if any"""
        macros = self.device_manager.connector.lrange(MessageEndpoints.pre_scan_macros(), 0, -1)
        for macro in macros:
            macro = macro.value.strip()
            func_name = self._get_func_name_from_macro(macro)
            exec(macro)
            eval(func_name)(self.device_manager.devices, self)

    def initialize(self):
        self.run_pre_scan_macros()

    def _check_limits(self):
        logger.debug("check limits")
        for ii, dev in enumerate(self.scan_motors):
            low_limit, high_limit = self.device_manager.devices[dev].limits
            if low_limit >= high_limit:
                # if both limits are equal or low > high, no restrictions ought to be applied
                return
            for pos in self.positions:
                pos_axis = pos[ii]
                if not low_limit <= pos_axis <= high_limit:
                    raise LimitError(
                        f"Target position {pos} for motor {dev} is outside of range: [{low_limit},"
                        f" {high_limit}]"
                    )

    def _get_scan_motors(self):
        if len(self.caller_args) == 0:
            return
        if self.arg_bundle_size.get("bundle"):
            self.scan_motors = list(self.caller_args.keys())
            return
        for motor in self.caller_args:
            if motor not in self.device_manager.devices:
                continue
            self.scan_motors.append(motor)

    def update_readout_priority(self):
        """update the readout priority for this request. Typically the monitored devices should also include the scan motors."""
        self.readout_priority["monitored"].extend(self.scan_motors)
        self.readout_priority["monitored"] = list(set(self.readout_priority["monitored"]))

    @abstractmethod
    def run(self):
        pass


class ScanBase(RequestBase, PathOptimizerMixin):
    """
    Base class for all scans. The following methods are called in the following order during the scan
    1. initialize
        - run_pre_scan_macros
    2. read_scan_motors
    3. prepare_positions
        - _calculate_positions
        - _optimize_trajectory
        - _set_position_offset
        - _check_limits
    4. open_scan
    5. stage
    6. run_baseline_reading
    7. pre_scan
    8. scan_core
    9. finalize
    10. unstage
    11. cleanup

    A subclass of ScanBase must implement the following methods:
    - _calculate_positions

    Attributes:
        scan_name (str): name of the scan
        scan_report_hint (str): hint for the scan report
        scan_type (str): scan type. Can be "step" or "fly"
        arg_input (list): list of scan argument types
        arg_bundle_size (dict):
            - bundle: number of arguments that are bundled together
            - min: minimum number of bundles
            - max: maximum number of bundles
        required_kwargs (list): list of required kwargs
        return_to_start_after_abort (bool): if True, the scan will return to the start position after an abort
    """

    scan_name = ""
    scan_report_hint = None
    scan_type = "step"
    arg_input = {"device": ScanArgType.DEVICE}
    arg_bundle_size = {"bundle": len(arg_input), "min": 1, "max": None}
    required_kwargs = ["required"]
    return_to_start_after_abort = True

    # perform pre-move action before the pre_scan trigger is sent
    pre_move = True

    # # synchronize primary readings with the scan motor readings. Set to False if the master device
    # # does not provide single readouts or are too fast to be synchronized with primary readings
    # enforce_sync = True

    def __init__(
        self,
        *args,
        device_manager: DeviceManagerBase = None,
        parameter: dict = None,
        exp_time: float = 0,
        readout_time: float = 0,
        acquisition_config: dict = None,
        settling_time: float = 0,
        relative: bool = False,
        burst_at_each_point: int = 1,
        frames_per_trigger: int = 1,
        optim_trajectory: Literal["corridor", None] = None,
        monitored: list = None,
        metadata: dict = None,
        **kwargs,
    ):
        super().__init__(
            *args,
            device_manager=device_manager,
            monitored=monitored,
            parameter=parameter,
            metadata=metadata,
            **kwargs,
        )
        self.DIID = 0
        self.point_id = 0
        self.exp_time = exp_time
        self.readout_time = readout_time
        self.acquisition_config = acquisition_config
        self.settling_time = settling_time
        self.relative = relative
        self.burst_at_each_point = burst_at_each_point
        self.frames_per_trigger = frames_per_trigger
        self.optim_trajectory = optim_trajectory
        self.burst_index = 0

        self.start_pos = np.repeat(0, len(self.scan_motors)).tolist()
        self.positions = []
        self.num_pos = None

        if self.scan_name == "":
            raise ValueError("scan_name cannot be empty")

        if acquisition_config is None or "default" not in acquisition_config:
            self.acquisition_config = {
                "default": {"exp_time": self.exp_time, "readout_time": self.readout_time}
            }

    @property
    def monitor_sync(self):
        """
        monitor_sync is a property that defines how monitored devices are synchronized.
        It can be either bec or the name of the device. If set to bec, the scan bundler
        will synchronize scan segments based on the bec triggered readouts. If set to a device name,
        the scan bundler will synchronize based on the readouts of the device, i.e. upon
        receiving a new readout of the device, cached monitored readings will be added
        to the scan segment.
        """
        return "bec"

    def read_scan_motors(self):
        """read the scan motors"""
        yield from self.stubs.read_and_wait(device=self.scan_motors, wait_group="scan_motor")

    @abstractmethod
    def _calculate_positions(self) -> None:
        """Calculate the positions"""
        pass

    def _optimize_trajectory(self):
        if not self.optim_trajectory:
            return
        if self.optim_trajectory == "corridor":
            self.positions = self.optimize_corridor(self.positions)
            return
        return

    def prepare_positions(self):
        """prepare the positions for the scan"""
        self._calculate_positions()
        self._optimize_trajectory()
        self.num_pos = len(self.positions) * self.burst_at_each_point
        yield from self._set_position_offset()
        self._check_limits()

    def open_scan(self):
        """open the scan"""
        positions = self.positions if isinstance(self.positions, list) else self.positions.tolist()
        yield from self.stubs.open_scan(
            scan_motors=self.scan_motors,
            readout_priority=self.readout_priority,
            num_pos=self.num_pos,
            positions=positions,
            scan_name=self.scan_name,
            scan_type=self.scan_type,
        )

    def stage(self):
        """call the stage procedure"""
        yield from self.stubs.stage()

    def run_baseline_reading(self):
        """perform a reading of all baseline devices"""
        yield from self.stubs.baseline_reading()

    def _set_position_offset(self):
        self.start_pos = []
        for dev in self.scan_motors:
            val = yield from self.stubs.send_rpc_and_wait(dev, "read")
            self.start_pos.append(val[dev].get("value"))
        if self.relative:
            self.positions += self.start_pos

    def close_scan(self):
        """close the scan"""
        yield from self.stubs.close_scan()

    def scan_core(self):
        """perform the scan core procedure"""
        for ind, pos in self._get_position():
            for self.burst_index in range(self.burst_at_each_point):
                yield from self._at_each_point(ind, pos)
            self.burst_index = 0

    def return_to_start(self):
        """return to the start position"""
        yield from self._move_and_wait(self.start_pos)

    def finalize(self):
        """finalize the scan"""
        yield from self.return_to_start()
        yield from self.stubs.wait(wait_type="read", group="primary", wait_group="readout_primary")
        yield from self.stubs.complete(device=None)

    def unstage(self):
        """call the unstage procedure"""
        yield from self.stubs.unstage()

    def cleanup(self):
        """call the cleanup procedure"""
        yield from self.close_scan()

    def _at_each_point(self, ind=None, pos=None):
        yield from self._move_and_wait(pos)
        if ind > 0:
            yield from self.stubs.wait(
                wait_type="read", group="primary", wait_group="readout_primary"
            )
        time.sleep(self.settling_time)
        yield from self.stubs.trigger(group="trigger", point_id=self.point_id)
        yield from self.stubs.wait(wait_type="trigger", group="trigger", wait_time=self.exp_time)
        yield from self.stubs.read(
            group="primary", wait_group="readout_primary", point_id=self.point_id
        )
        yield from self.stubs.wait(
            wait_type="read", group="scan_motor", wait_group="readout_primary"
        )

        self.point_id += 1

    def _move_and_wait(self, pos):
        if not isinstance(pos, list) and not isinstance(pos, np.ndarray):
            pos = [pos]
        if len(pos) == 0:
            return
        for ind, val in enumerate(self.scan_motors):
            yield from self.stubs.set(device=val, value=pos[ind], wait_group="scan_motor")

        yield from self.stubs.wait(wait_type="move", group="scan_motor", wait_group="scan_motor")

    def _get_position(self):
        for ind, pos in enumerate(self.positions):
            yield (ind, pos)

    def scan_report_instructions(self):
        yield None

    def pre_scan(self):
        """
        pre scan procedure. This method is called before the scan_core method and can be used to
        perform additional tasks before the scan is started. This
        """
        if self.pre_move and len(self.positions) > 0:
            yield from self._move_and_wait(self.positions[0])
        yield from self.stubs.pre_scan()

    def run(self):
        """run the scan. This method is called by the scan server and is the main entry point for the scan."""
        self.initialize()
        yield from self.read_scan_motors()
        yield from self.prepare_positions()
        yield from self.scan_report_instructions()
        yield from self.open_scan()
        yield from self.stage()
        yield from self.run_baseline_reading()
        yield from self.pre_scan()
        yield from self.scan_core()
        yield from self.finalize()
        yield from self.unstage()
        yield from self.cleanup()

    @classmethod
    def scan(cls, *args, **kwargs):
        scan = cls(args, **kwargs)
        yield from scan.run()


class SyncFlyScanBase(ScanBase, ABC):
    """
    Fly scan base class for all synchronous fly scans. A synchronous fly scan is a scan where the flyer is
    synced with the monitored devices.
    Classes inheriting from SyncFlyScanBase must at least implement the scan_core method and the monitor_sync property.
    """

    scan_type = "fly"
    pre_move = False

    def _get_flyer_status(self) -> list:
        flyer = self.scan_motors[0]
        connector = self.device_manager.connector

        pipe = connector.pipeline()
        connector.lrange(
            MessageEndpoints.device_req_status_container(self.metadata["RID"]), 0, -1, pipe
        )
        connector.get(MessageEndpoints.device_readback(flyer), pipe)
        return connector.execute_pipeline(pipe)

    @abstractmethod
    def scan_core(self):
        """perform the scan core procedure"""
        ############################################
        # Example of how to kickoff and wait for a flyer:
        ############################################

        # yield from self.stubs.kickoff(device=self.scan_motors[0], parameter=self.caller_kwargs)
        # yield from self.stubs.complete(device=self.scan_motors[0])
        # target_diid = self.DIID - 1

        # while True:
        #     status = self.stubs.get_req_status(
        #         device=self.scan_motors[0], RID=self.metadata["RID"], DIID=target_diid
        #     )
        #     progress = self.stubs.get_device_progress(
        #         device=self.scan_motors[0], RID=self.metadata["RID"]
        #     )
        #     if progress:
        #         self.num_pos = progress
        #     if status:
        #         break
        #     time.sleep(1)

    @property
    @abstractmethod
    def monitor_sync(self) -> str:
        """
        monitor_sync is the flyer that will be used to synchronize the monitor readings in the scan bundler.
        The return value should be the name of the flyer device.
        """

    def _calculate_positions(self) -> None:
        pass

    def read_scan_motors(self):
        yield None

    def prepare_positions(self):
        yield None


class AsyncFlyScanBase(SyncFlyScanBase):
    """
    Fly scan base class for all asynchronous fly scans. An asynchronous fly scan is a scan where the flyer is
    not synced with the monitored devices.
    Classes inheriting from AsyncFlyScanBase must at least implement the scan_core method.
    """

    @property
    def monitor_sync(self):
        return "bec"


class ScanStub(RequestBase):
    pass


class OpenScanDef(ScanStub):
    scan_name = "open_scan_def"
    scan_report_hint = None
    arg_input = {}
    arg_bundle_size = {"bundle": len(arg_input), "min": 0, "max": 0}

    def run(self):
        yield from self.stubs.open_scan_def()


class CloseScanDef(ScanStub):
    scan_name = "close_scan_def"
    scan_report_hint = "table"
    arg_input = {}
    arg_bundle_size = {"bundle": len(arg_input), "min": 0, "max": 0}

    def run(self):
        yield from self.stubs.close_scan_def()


class CloseScanGroup(ScanStub):
    scan_name = "close_scan_group"
    arg_input = {}
    arg_bundle_size = {"bundle": len(arg_input), "min": 0, "max": 0}

    def run(self):
        yield from self.stubs.close_scan_group()


class DeviceRPC(ScanStub):
    scan_name = "device_rpc"
    arg_input = [ScanArgType.DEVICE, ScanArgType.STR, ScanArgType.LIST, ScanArgType.DICT]
    arg_bundle_size = {"bundle": len(arg_input), "min": 1, "max": 1}
    scan_report_hint = None

    def _get_scan_motors(self):
        pass

    def run(self):
        # different to calling self.device_rpc, this procedure will not wait for a reply and therefore not check any errors.
        yield from self.stubs.rpc(device=self.parameter.get("device"), parameter=self.parameter)


class Move(RequestBase):
    scan_name = "mv"
    arg_input = {"device": ScanArgType.DEVICE, "target": ScanArgType.FLOAT}
    arg_bundle_size = {"bundle": len(arg_input), "min": 1, "max": None}
    scan_report_hint = None
    required_kwargs = ["relative"]

    def __init__(self, *args, relative=False, **kwargs):
        """
        Move device(s) to an absolute position
        Args:
            *args (Device, float): pairs of device / position arguments
            relative (bool): if True, move relative to current position

        Returns:
            ScanReport

        Examples:
            >>> scans.mv(dev.samx, 1, dev.samy,2)
        """
        super().__init__(**kwargs)
        self.relative = relative
        self.start_pos = np.repeat(0, len(self.scan_motors)).tolist()

    def _calculate_positions(self):
        self.positions = np.asarray([[val[0] for val in self.caller_args.values()]], dtype=float)

    def _at_each_point(self, pos=None):
        for ii, motor in enumerate(self.scan_motors):
            yield from self.stubs.set(
                device=motor,
                value=self.positions[0][ii],
                wait_group="scan_motor",
                metadata={"response": True},
            )

    def cleanup(self):
        pass

    def _set_position_offset(self):
        self.start_pos = []
        for dev in self.scan_motors:
            val = yield from self.stubs.send_rpc_and_wait(dev, "read")
            self.start_pos.append(val[dev].get("value"))
        if not self.relative:
            return
        self.positions += self.start_pos

    def prepare_positions(self):
        self._calculate_positions()
        yield from self._set_position_offset()
        self._check_limits()

    def scan_report_instructions(self):
        if not self.scan_report_hint:
            yield None
            return
        yield from self.stubs.scan_report_instruction(
            {
                "readback": {
                    "RID": self.metadata["RID"],
                    "devices": self.scan_motors,
                    "start": self.start_pos,
                    "end": self.positions[0],
                }
            }
        )

    def run(self):
        self.initialize()
        yield from self.prepare_positions()
        yield from self.scan_report_instructions()
        yield from self._at_each_point()


class UpdatedMove(Move):
    """
    Move device(s) to an absolute position and show live updates. This is a blocking call. For non-blocking use Move.
    Args:
        *args (Device, float): pairs of device / position arguments
        relative (bool): if True, move relative to current position

    Returns:
        ScanReport

    Examples:
        >>> scans.umv(dev.samx, 1, dev.samy,2)
    """

    scan_name = "umv"
    scan_report_hint = "readback"

    def _at_each_point(self, pos=None):
        for ii, motor in enumerate(self.scan_motors):
            yield from self.stubs.set(
                device=motor, value=self.positions[0][ii], wait_group="scan_motor"
            )

        for motor in self.scan_motors:
            yield from self.stubs.wait(wait_type="move", device=motor, wait_group="scan_motor")


class Scan(ScanBase):
    scan_name = "grid_scan"
    scan_report_hint = "table"
    arg_input = {
        "device": ScanArgType.DEVICE,
        "start": ScanArgType.FLOAT,
        "stop": ScanArgType.FLOAT,
        "steps": ScanArgType.INT,
    }
    arg_bundle_size = {"bundle": len(arg_input), "min": 2, "max": None}
    required_kwargs = ["relative"]

    def __init__(
        self,
        *args,
        exp_time: float = 0,
        settling_time: float = 0,
        relative: bool = False,
        burst_at_each_point: int = 1,
        **kwargs,
    ):
        """
        Scan two motors in a grid.

        Args:
            *args (Device, float, float, int): pairs of device / start / stop / steps arguments
            exp_time (float): exposure time in seconds. Default is 0.
            settling_time (float): settling time in seconds. Default is 0.
            relative (bool): if True, the motors will be moved relative to their current position. Default is False.
            burst_at_each_point (int): number of exposures at each point. Default is 1.

        Returns:
            ScanReport

        Examples:
            >>> scans.grid_scan(dev.motor1, -5, 5, 10, dev.motor2, -5, 5, 10, exp_time=0.1, relative=True)

        """
        super().__init__(**kwargs)
        self.relative = relative
        self.exp_time = exp_time
        self.settling_time = settling_time
        self.burst_at_each_point = burst_at_each_point
        self.axis = []

    def _calculate_positions(self):
        for _, val in self.caller_args.items():
            self.axis.append(np.linspace(val[0], val[1], val[2]))
        if len(self.axis) > 1:
            self.positions = get_2D_raster_pos(self.axis)
        else:
            self.positions = np.vstack(tuple(self.axis)).T


class FermatSpiralScan(ScanBase):
    scan_name = "fermat_scan"
    scan_report_hint = "table"
    required_kwargs = ["step", "relative"]
    arg_input = {
        "device": ScanArgType.DEVICE,
        "start": ScanArgType.FLOAT,
        "stop": ScanArgType.FLOAT,
    }
    arg_bundle_size = {"bundle": len(arg_input), "min": 2, "max": 2}

    def __init__(
        self,
        *args,
        step: float = 0.1,
        exp_time: float = 0,
        settling_time: float = 0,
        relative: bool = False,
        burst_at_each_point: int = 1,
        spiral_type: float = 0,
        optim_trajectory: Literal["corridor", None] = None,
        **kwargs,
    ):
        """
        A scan following Fermat's spiral.

        Args:
            *args: pairs of device / start position / end position arguments
            step (float): step size in motor units. Default is 0.1.
            exp_time (float): exposure time in seconds. Default is 0.
            settling_time (float): settling time in seconds. Default is 0.
            relative (bool): if True, the motors will be moved relative to their current position. Default is False.
            burst_at_each_point (int): number of exposures at each point. Default is 1.
            spiral_type (float): type of spiral to use. Default is 0.
            optim_trajectory (str): trajectory optimization method. Default is None. Options are "corridor" and "none".

        Returns:
            ScanReport

        Examples:
            >>> scans.fermat_scan(dev.motor1, -5, 5, dev.motor2, -5, 5, step=0.5, exp_time=0.1, relative=True, optim_trajectory="corridor")

        """
        super().__init__(**kwargs)
        self.axis = []
        self.step = step
        self.exp_time = exp_time
        self.settling_time = settling_time
        self.relative = relative
        self.burst_at_each_point = burst_at_each_point
        self.spiral_type = spiral_type
        self.optim_trajectory = optim_trajectory

    def _calculate_positions(self):
        params = list(self.caller_args.values())
        self.positions = get_fermat_spiral_pos(
            params[0][0],
            params[0][1],
            params[1][0],
            params[1][1],
            step=self.step,
            spiral_type=self.spiral_type,
            center=False,
        )


class RoundScan(ScanBase):
    scan_name = "round_scan"
    scan_report_hint = "table"
    required_kwargs = ["relative"]
    arg_input = {
        "motor_1": ScanArgType.DEVICE,
        "motor_2": ScanArgType.DEVICE,
        "inner_ring": ScanArgType.FLOAT,
        "outer_ring": ScanArgType.FLOAT,
        "number_of_rings": ScanArgType.INT,
        "number_of_positions_in_first_ring": ScanArgType.INT,
    }
    arg_bundle_size = {"bundle": len(arg_input), "min": 1, "max": 1}

    def __init__(
        self,
        motor_1,
        motor2,
        inner_ring: float,
        outer_ring: float,
        number_of_rings: int,
        pos_in_first_ring: int,
        *args,
        relative: bool = False,
        burst_at_each_point: int = 1,
        **kwargs,
    ):
        """
        A scan following a round shell-like pattern.

        Args:
            *args: motor1, motor2, inner ring, outer ring, number of rings, number of positions in the first ring
            relative (bool): if True, the motors will be moved relative to their current position. Default is False.
            burst_at_each_point (int): number of exposures at each point. Default is 1.

        Returns:
            ScanReport

        Examples:
            >>> scans.round_scan(dev.motor1, dev.motor2, 0, 25, 5, 3, exp_time=0.1, relative=True)

        """
        super().__init__(**kwargs)
        self.relative = relative
        self.burst_at_each_point = burst_at_each_point
        self.axis = []

    def _get_scan_motors(self):
        caller_args = list(self.caller_args.items())[0]
        self.scan_motors = [caller_args[0], caller_args[1][0]]

    def _calculate_positions(self):
        params = list(self.caller_args.values())[0]
        self.positions = get_round_scan_positions(
            r_in=params[1], r_out=params[2], nr=params[3], nth=params[4]
        )


class ContLineScan(ScanBase):
    scan_name = "cont_line_scan"
    scan_report_hint = "table"
    required_kwargs = ["steps", "relative"]
    arg_input = {
        "device": ScanArgType.DEVICE,
        "start": ScanArgType.FLOAT,
        "stop": ScanArgType.FLOAT,
    }
    arg_bundle_size = {"bundle": len(arg_input), "min": 1, "max": 1}
    scan_type = "step"

    def __init__(
        self,
        *args,
        exp_time: float = 0,
        steps: int = 10,
        relative: bool = False,
        offset: float = 100,
        burst_at_each_point: int = 1,
        **kwargs,
    ):
        """
        A line scan for one or more motors.

        Args:
            *args (Device, float, float): pairs of device / start position / end position
            exp_time (float): exposure time in seconds. Default is 0.
            steps (int): number of steps. Default is 10.
            relative (bool): if True, the motors will be moved relative to their current position. Default is False.
            burst_at_each_point (int): number of exposures at each point. Default is 1.
            offset (float): offset in motor units. Default is 100.

        Returns:
            ScanReport

        Examples:
            >>> scans.cont_line_scan(dev.motor1, -5, 5, steps=10, exp_time=0.1, relative=True)

        """
        super().__init__(**kwargs)
        self.axis = []
        self.exp_time = exp_time
        self.steps = steps
        self.relative = relative
        self.burst_at_each_point = burst_at_each_point
        self.offset = offset

    def _calculate_positions(self) -> None:
        for _, val in self.caller_args.items():
            ax_pos = np.linspace(val[0], val[1], self.steps)
            self.axis.append(ax_pos)
        self.positions = np.array(list(zip(*self.axis)))

    def _at_each_point(self):
        yield from self.stubs.trigger(group="trigger", point_id=self.point_id)
        yield from self.stubs.read(group="primary", wait_group="primary", point_id=self.point_id)
        self.point_id += 1

    def scan_core(self):
        yield from self._move_and_wait(self.positions[0] - self.offset)
        # send the slow motor on its way
        yield from self.stubs.set(
            device=self.scan_motors[0], value=self.positions[-1][0], wait_group="scan_motor"
        )

        while self.point_id < len(self.positions[:]):
            cont_motor_positions = self.device_manager.devices[self.scan_motors[0]].readback()

            if not cont_motor_positions:
                continue

            cont_motor_positions = cont_motor_positions.get("value")
            logger.debug(f"Current position of {self.scan_motors[0]}: {cont_motor_positions}")
            if np.isclose(cont_motor_positions, self.positions[self.point_id][0], atol=0.5):
                logger.debug(f"reading point {self.point_id}")
                yield from self._at_each_point()
                continue
            if cont_motor_positions > self.positions[self.point_id][0]:
                raise ScanAbortion(f"Skipped point {self.point_id + 1}")


class RoundScanFlySim(SyncFlyScanBase):
    scan_name = "round_scan_fly"
    scan_report_hint = "table"
    scan_type = "fly"
    pre_move = False
    required_kwargs = ["relative"]
    arg_input = {
        "flyer": ScanArgType.DEVICE,
        "inner_ring": ScanArgType.FLOAT,
        "outer_ring": ScanArgType.FLOAT,
        "number_of_rings": ScanArgType.INT,
        "number_of_positions_in_first_ring": ScanArgType.INT,
    }
    arg_bundle_size = {"bundle": len(arg_input), "min": 1, "max": 1}

    def __init__(self, *args, **kwargs):
        """
        A fly scan following a round shell-like pattern.

        Args:
            *args: motor1, motor2, inner ring, outer ring, number of rings, number of positions in the first ring
            relative: Start from an absolute or relative position
            burst: number of acquisition per point

        Returns:
            ScanReport

        Examples:
            >>> scans.round_scan_fly(dev.flyer_sim, 0, 50, 5, 3, exp_time=0.1, relative=True)

        """
        super().__init__(**kwargs)
        self.axis = []

    def _get_scan_motors(self):
        self.scan_motors = []
        self.flyer = list(self.caller_args.keys())[0]

    @property
    def monitor_sync(self):
        return self.flyer

    def prepare_positions(self):
        self._calculate_positions()
        self.num_pos = len(self.positions) * self.burst_at_each_point
        self._check_limits()
        yield None

    def finalize(self):
        yield

    def _calculate_positions(self):
        params = list(self.caller_args.values())[0]
        self.positions = get_round_scan_positions(
            r_in=params[0], r_out=params[1], nr=params[2], nth=params[3]
        )

    def scan_core(self):
        yield from self.stubs.kickoff(
            device=self.flyer,
            parameter={
                "num_pos": self.num_pos,
                "positions": self.positions.tolist(),
                "exp_time": self.exp_time,
            },
        )
        target_DIID = self.DIID - 1

        while True:
            yield from self.stubs.read_and_wait(group="primary", wait_group="readout_primary")
            status = self.device_manager.connector.get(MessageEndpoints.device_status(self.flyer))
            if status:
                device_is_idle = status.content.get("status", 1) == 0
                matching_RID = self.metadata.get("RID") == status.metadata.get("RID")
                matching_DIID = target_DIID == status.metadata.get("DIID")
                if device_is_idle and matching_RID and matching_DIID:
                    break

            time.sleep(1)
            logger.debug("reading monitors")


class RoundROIScan(ScanBase):
    scan_name = "round_roi_scan"
    scan_report_hint = "table"
    required_kwargs = ["dr", "nth", "relative"]
    arg_input = {
        "motor_1": ScanArgType.DEVICE,
        "motor_2": ScanArgType.DEVICE,
        "width_1": ScanArgType.FLOAT,
        "width_2": ScanArgType.FLOAT,
    }
    arg_bundle_size = {"bundle": len(arg_input), "min": 1, "max": 1}

    def __init__(
        self,
        *args,
        dr: float = 1,
        nth: int = 5,
        exp_time: float = 0,
        relative: bool = False,
        burst_at_each_point: int = 1,
        **kwargs,
    ):
        """
        A scan following a round-roi-like pattern.

        Args:
            *args: motor1, width for motor1, motor2, width for motor2,
            dr (float): shell width. Default is 1.
            nth (int): number of points in the first shell. Default is 5.
            exp_time (float): exposure time in seconds. Default is 0.
            relative (bool): Start from an absolute or relative position. Default is False.
            burst_at_each_point (int): number of acquisition per point. Default is 1.

        Returns:
            ScanReport

        Examples:
            >>> scans.round_roi_scan(dev.motor1, 20, dev.motor2, 20, dr=2, nth=3, exp_time=0.1, relative=True)

        """
        super().__init__(**kwargs)
        self.axis = []
        self.dr = dr
        self.nth = nth
        self.exp_time = exp_time
        self.relative = relative
        self.burst_at_each_point = burst_at_each_point

    def _calculate_positions(self) -> None:
        params = list(self.caller_args.values())
        self.positions = get_round_roi_scan_positions(
            lx=params[0][0], ly=params[1][0], dr=self.dr, nth=self.nth
        )


class ListScan(ScanBase):
    scan_name = "list_scan"
    scan_report_hint = "table"
    required_kwargs = ["relative"]
    arg_input = {"device": ScanArgType.DEVICE, "positions": ScanArgType.LIST}
    arg_bundle_size = {"bundle": len(arg_input), "min": 1, "max": None}

    def __init__(self, *args, parameter: dict = None, **kwargs):
        """
        A scan following the positions specified in a list.
        Please note that all lists must be of equal length.

        Args:
            *args: pairs of motors and position lists
            relative: Start from an absolute or relative position
            burst: number of acquisition per point

        Returns:
            ScanReport

        Examples:
            >>> scans.list_scan(dev.motor1, [0,1,2,3,4], dev.motor2, [4,3,2,1,0], exp_time=0.1, relative=True)

        """
        super().__init__(parameter=parameter, **kwargs)
        self.axis = []
        if len(set(len(entry[0]) for entry in self.caller_args.values())) != 1:
            raise ValueError("All position lists must be of equal length.")

    def _calculate_positions(self):
        self.positions = np.vstack(tuple(self.caller_args.values())).T.tolist()


class TimeScan(ScanBase):
    scan_name = "time_scan"
    scan_report_hint = "table"
    required_kwargs = ["points", "interval"]
    arg_input = {}
    arg_bundle_size = {"bundle": len(arg_input), "min": None, "max": None}

    def __init__(
        self,
        points: int,
        interval: float,
        exp_time: float = 0,
        burst_at_each_point: int = 1,
        **kwargs,
    ):
        """
        Trigger and readout devices at a fixed interval.
        Note that the interval time cannot be less than the exposure time.
        The effective "sleep" time between points is
            sleep_time = interval - exp_time

        Args:
            points: number of points
            interval: time interval between points
            exp_time: exposure time in s
            burst: number of acquisition per point

        Returns:
            ScanReport

        Examples:
            >>> scans.time_scan(points=10, interval=1.5, exp_time=0.1, relative=True)

        """
        super().__init__(**kwargs)
        self.axis = []
        self.points = points
        self.exp_time = exp_time
        self.burst_at_each_point = burst_at_each_point
        self.interval = interval
        self.interval -= self.exp_time

    def _calculate_positions(self) -> None:
        pass

    def prepare_positions(self):
        self.num_pos = self.points
        yield None

    def _at_each_point(self, ind=None, pos=None):
        if ind > 0:
            yield from self.stubs.wait(
                wait_type="read", group="primary", wait_group="readout_primary"
            )
        yield from self.stubs.trigger(group="trigger", point_id=self.point_id)
        yield from self.stubs.wait(wait_type="trigger", group="trigger", wait_time=self.exp_time)
        yield from self.stubs.read(
            group="primary", wait_group="readout_primary", point_id=self.point_id
        )
        yield from self.stubs.wait(wait_type="trigger", group="trigger", wait_time=self.interval)
        self.point_id += 1

    def scan_core(self):
        for ind in range(self.num_pos):
            yield from self._at_each_point(ind)


class MonitorScan(ScanBase):
    scan_name = "monitor_scan"
    scan_report_hint = "table"
    required_kwargs = ["relative"]
    arg_input = {
        "device": ScanArgType.DEVICE,
        "start": ScanArgType.FLOAT,
        "stop": ScanArgType.FLOAT,
    }
    arg_bundle_size = {"bundle": len(arg_input), "min": 1, "max": 1}
    scan_type = "fly"

    def __init__(self, device, start: float, stop: float, *args, relative: bool = False, **kwargs):
        """
        Readout all primary devices at each update of the monitored device.

        Args:
            device (Device): monitored device
            start (float): start position of the monitored device
            stop (float): stop position of the monitored device

        Returns:
            ScanReport

        Examples:
            >>> scans.monitor_scan(dev.motor1, -5, 5, exp_time=0.1, relative=True)

        """
        self.device = device
        self.start = start
        self.stop = stop
        super().__init__(**kwargs)
        self.axis = []
        self.relative = relative

    def _get_scan_motors(self):
        self.scan_motors = [self.device]
        self.flyer = self.device

    @property
    def monitor_sync(self):
        return self.flyer

    def _calculate_positions(self) -> None:
        self.positions = np.vstack(tuple(self.caller_args.values())).T.tolist()

    def prepare_positions(self):
        self._calculate_positions()
        self.num_pos = 0
        yield from self._set_position_offset()
        self._check_limits()

    def _get_flyer_status(self) -> list:
        connector = self.device_manager.connector

        pipe = connector.pipeline()
        connector.lrange(
            MessageEndpoints.device_req_status_container(self.metadata["RID"]), 0, -1, pipe
        )
        connector.get(MessageEndpoints.device_readback(self.flyer), pipe)
        return connector.execute_pipeline(pipe)

    def scan_core(self):
        yield from self.stubs.set(
            device=self.flyer, value=self.positions[0][0], wait_group="scan_motor"
        )
        yield from self.stubs.wait(wait_type="move", device=self.flyer, wait_group="scan_motor")

        # send the slow motor on its way
        yield from self.stubs.set(
            device=self.flyer,
            value=self.positions[1][0],
            wait_group="scan_motor",
            metadata={"response": True},
        )

        while True:
            move_completed, readback = self._get_flyer_status()

            if move_completed:
                break

            if not readback:
                continue
            readback = readback.content["signals"]
            yield from self.stubs.publish_data_as_read(
                device=self.flyer, data=readback, point_id=self.point_id
            )
            self.point_id += 1
            self.num_pos += 1


class Acquire(ScanBase):
    scan_name = "acquire"
    scan_report_hint = "table"
    required_kwargs = []
    arg_input = {}
    arg_bundle_size = {"bundle": len(arg_input), "min": None, "max": None}

    def __init__(self, *args, exp_time: float = 0, burst_at_each_point: int = 1, **kwargs):
        """
        A simple acquisition at the current position.

        Args:
            burst: number of acquisition per point

        Returns:
            ScanReport

        Examples:
            >>> scans.acquire(exp_time=0.1, relative=True)

        """
        super().__init__(**kwargs)
        self.exp_time = exp_time
        self.burst_at_each_point = burst_at_each_point
        self.axis = []

    def _calculate_positions(self) -> None:
        self.num_pos = self.burst_at_each_point

    def prepare_positions(self):
        self._calculate_positions()

    def _at_each_point(self, ind=None, pos=None):
        if ind > 0:
            yield from self.stubs.wait(
                wait_type="read", group="primary", wait_group="readout_primary"
            )
        yield from self.stubs.trigger(group="trigger", point_id=self.point_id)
        yield from self.stubs.wait(wait_type="trigger", group="trigger", wait_time=self.exp_time)
        yield from self.stubs.read(
            group="primary", wait_group="readout_primary", point_id=self.point_id
        )
        self.point_id += 1

    def scan_core(self):
        for self.burst_index in range(self.burst_at_each_point):
            yield from self._at_each_point(self.burst_index)
        self.burst_index = 0

    def run(self):
        self.initialize()
        self.prepare_positions()
        yield from self.open_scan()
        yield from self.stage()
        yield from self.run_baseline_reading()
        yield from self.pre_scan()
        yield from self.scan_core()
        yield from self.finalize()
        yield from self.unstage()
        yield from self.cleanup()


class LineScan(ScanBase):
    scan_name = "line_scan"
    scan_report_hint = "table"
    required_kwargs = ["steps", "relative"]
    arg_input = {
        "device": ScanArgType.DEVICE,
        "start": ScanArgType.FLOAT,
        "stop": ScanArgType.FLOAT,
    }
    arg_bundle_size = {"bundle": len(arg_input), "min": 1, "max": None}

    def __init__(
        self,
        *args,
        exp_time: float = 0,
        steps: int = None,
        relative: bool = False,
        burst_at_each_point: int = 1,
        **kwargs,
    ):
        """
        A line scan for one or more motors.

        Args:
            *args (Device, float, float): pairs of device / start position / end position
            exp_time (float): exposure time in s. Default: 0
            steps (int): number of steps. Default: 10
            relative (bool): if True, the start and end positions are relative to the current position. Default: False
            burst_at_each_point (int): number of acquisition per point. Default: 1

        Returns:
            ScanReport

        Examples:
            >>> scans.line_scan(dev.motor1, -5, 5, dev.motor2, -5, 5, steps=10, exp_time=0.1, relative=True)

        """
        super().__init__(**kwargs)
        self.exp_time = exp_time
        self.steps = steps
        self.relative = relative
        self.burst_at_each_point = burst_at_each_point
        self.axis = []

    def _calculate_positions(self) -> None:
        for _, val in self.caller_args.items():
            ax_pos = np.linspace(val[0], val[1], self.steps)
            self.axis.append(ax_pos)
        self.positions = np.array(list(zip(*self.axis)))


class ScanComponent(ScanBase):
    pass


class OpenInteractiveScan(ScanComponent):
    scan_name = "open_interactive_scan"
    scan_report_hint = ""
    required_kwargs = []
    arg_input = {"device": ScanArgType.DEVICE}
    arg_bundle_size = {"bundle": len(arg_input), "min": 1, "max": None}

    def __init__(self, *args, **kwargs):
        """
        An interactive scan for one or more motors.

        Args:
            *args: devices
            exp_time: exposure time in s
            steps: number of steps (please note: 5 steps == 6 positions)
            relative: Start from an absolute or relative position
            burst: number of acquisition per point

        Returns:
            ScanReport

        Examples:
            >>> scans.open_interactive_scan(dev.motor1, dev.motor2, exp_time=0.1)

        """
        super().__init__(**kwargs)
        self.axis = []

    def _calculate_positions(self):
        pass

    def _get_scan_motors(self):
        caller_args = list(self.caller_args.keys())
        self.scan_motors = caller_args

    def run(self):
        yield from self.stubs.open_scan_def()
        self.initialize()
        yield from self.read_scan_motors()
        yield from self.open_scan()
        yield from self.stage()
        yield from self.run_baseline_reading()


class AddInteractiveScanPoint(ScanComponent):
    scan_name = "interactive_scan_trigger"
    scan_report_hint = ""
    arg_input = {"device": ScanArgType.DEVICE}
    arg_bundle_size = {"bundle": len(arg_input), "min": 1, "max": None}

    def __init__(self, *args, **kwargs):
        """
        An interactive scan for one or more motors.

        Args:
            *args: devices
            exp_time: exposure time in s
            steps: number of steps (please note: 5 steps == 6 positions)
            relative: Start from an absolute or relative position
            burst: number of acquisition per point

        Returns:
            ScanReport

        Examples:
            >>> scans.interactive_scan_trigger()

        """
        super().__init__(**kwargs)
        self.axis = []

    def _calculate_positions(self):
        pass

    def _get_scan_motors(self):
        self.scan_motors = list(self.caller_args.keys())

    def _at_each_point(self, ind=None, pos=None):
        yield from self.stubs.trigger(group="trigger", point_id=self.point_id)
        yield from self.stubs.wait(wait_type="trigger", group="trigger", wait_time=self.exp_time)
        yield from self.stubs.read_and_wait(
            group="primary", wait_group="readout_primary", point_id=self.point_id
        )
        self.point_id += 1

    def run(self):
        yield from self.open_scan()
        yield from self._at_each_point()
        yield from self.close_scan()


class CloseInteractiveScan(ScanComponent):
    scan_name = "close_interactive_scan"
    scan_report_hint = ""
    arg_input = {}
    arg_bundle_size = {"bundle": len(arg_input), "min": None, "max": None}

    def __init__(self, *args, **kwargs):
        """
        An interactive scan for one or more motors.

        Args:
            *args: devices
            exp_time: exposure time in s
            steps: number of steps (please note: 5 steps == 6 positions)
            relative: Start from an absolute or relative position
            burst: number of acquisition per point

        Returns:
            ScanReport

        Examples:
            >>> scans.close_interactive_scan(dev.motor1, dev.motor2, exp_time=0.1)

        """
        super().__init__(**kwargs)
        self.axis = []

    def _calculate_positions(self):
        pass

    def run(self):
        yield from self.finalize()
        yield from self.unstage()
        yield from self.cleanup()
        yield from self.stubs.close_scan_def()



import numpy as np
import pytest

from bec_lib.serialization import json_ext, msgpack


@pytest.fixture(params=[json_ext, msgpack])
def serializer(request):
    yield request.param


@pytest.mark.parametrize(
    "data",
    [
        {"a": 1, "b": 2},
        "hello",
        1,
        1.0,
        [1, 2, 3],
        np.array([1, 2, 3]),
        {1, 2, 3},
        {
            "hroz": {
                "hroz": {"value": 0, "timestamp": 1708336264.5731058},
                "hroz_setpoint": {"value": 0, "timestamp": 1708336264.573121},
            }
        },
    ],
)
def test_serialize(serializer, data):
    res = serializer.loads(serializer.dumps(data)) == data
    assert all(res) if isinstance(data, np.ndarray) else res



import datetime
import sys
import time
from unittest import mock

import pytest

from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_lib.queue_items import QueueItem
from bec_lib.scan_data import ScanData
from bec_lib.scan_items import ScanItem
from bec_lib.scan_manager import ScanManager
from bec_lib.tests.utils import ConnectorMock

# pylint: disable=missing-function-docstring


@pytest.mark.parametrize(
    "queue_msg",
    [
        messages.ScanQueueStatusMessage(
            queue={
                "primary": {
                    "info": [
                        {
                            "queue_id": "7c15c9a2-71d4-4f2a-91a7-c4a63088fa38",
                            "scan_id": ["bfa582aa-f9cd-4258-ab5d-3e5d54d3dde5"],
                            "is_scan": [True],
                            "request_blocks": [
                                {
                                    "msg": b"\x84\xa8msg_type\xa4scan\xa7content\x83\xa9scan_type\xabfermat_scan\xa9parameter\x82\xa4args\x82\xa4samx\x92\xfe\x02\xa4samy\x92\xfe\x02\xa6kwargs\x83\xa4step\xcb?\xf8\x00\x00\x00\x00\x00\x00\xa8exp_time\xcb?\x94z\xe1G\xae\x14{\xa8relative\xc3\xa5queue\xa7primary\xa8metadata\x81\xa3RID\xd9$cd8fc68f-fe65-4031-9a37-e0e7ba9df542\xa7version\xcb?\xf0\x00\x00\x00\x00\x00\x00",
                                    "RID": "cd8fc68f-fe65-4031-9a37-e0e7ba9df542",
                                    "scan_motors": ["samx", "samy"],
                                    "is_scan": True,
                                    "scan_number": 25,
                                    "scan_id": "bfa582aa-f9cd-4258-ab5d-3e5d54d3dde5",
                                    "metadata": {"RID": "cd8fc68f-fe65-4031-9a37-e0e7ba9df542"},
                                    "content": {
                                        "scan_type": "fermat_scan",
                                        "parameter": {
                                            "args": {"samx": [-2, 2], "samy": [-2, 2]},
                                            "kwargs": {
                                                "step": 1.5,
                                                "exp_time": 0.02,
                                                "relative": True,
                                            },
                                        },
                                        "queue": "primary",
                                    },
                                }
                            ],
                            "scan_number": [25],
                            "status": "PENDING",
                            "active_request_block": None,
                        }
                    ],
                    "status": "RUNNING",
                }
            }
        )
    ],
)
def test_update_with_queue_status(queue_msg):
    scan_manager = ScanManager(ConnectorMock(""))
    scan_manager.connector._get_buffer[MessageEndpoints.scan_queue_status().endpoint] = queue_msg
    scan_manager.update_with_queue_status(queue_msg)
    assert (
        scan_manager.scan_storage.find_scan_by_ID("bfa582aa-f9cd-4258-ab5d-3e5d54d3dde5")
        is not None
    )


def test_scan_item_to_pandas():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_item = ScanItem(scan_manager, "queue_id", [1], ["scan_id"], "status")
    scan_item.data = ScanData()
    data = {
        0: messages.ScanMessage(
            point_id=0, scan_id="scan_id", data={"samx": {"samx": {"value": 1, "timestamp": 0}}}
        ),
        1: messages.ScanMessage(
            point_id=1, scan_id="scan_id", data={"samx": {"samx": {"value": 2, "timestamp": 0}}}
        ),
        2: messages.ScanMessage(
            point_id=2, scan_id="scan_id", data={"samx": {"samx": {"value": 3, "timestamp": 0}}}
        ),
    }
    for ii, msg in data.items():
        scan_item.data.set(ii, msg)

    df = scan_item.to_pandas()
    assert df["samx"]["samx"]["value"].tolist() == [1, 2, 3]
    assert df["samx"]["samx"]["timestamp"].tolist() == [0, 0, 0]


def test_scan_item_to_pandas_empty_data():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_item = ScanItem(scan_manager, "queue_id", [1], ["scan_id"], "status")
    scan_item.data = ScanData()

    df = scan_item.to_pandas()
    assert df.empty


def test_scan_item_to_pandas_raises_without_pandas_installed():
    """Test that to_pandas raises an ImportError if pandas is not installed."""
    scan_manager = ScanManager(ConnectorMock(""))
    scan_item = ScanItem(scan_manager, "queue_id", [1], ["scan_id"], "status")

    with mock.patch.dict("sys.modules"):
        del sys.modules["pandas"]
        with pytest.raises(ImportError):
            scan_item.to_pandas()


def test_scan_item_str():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_item = ScanItem(scan_manager, "queue_id", [1], ["scan_id"], "status")
    start_time = "Fri Jun 23 15:11:06 2023"
    # convert to datetime string to timestamp
    scan_item.start_time = time.mktime(
        datetime.datetime.strptime(start_time, "%a %b %d %H:%M:%S %Y").timetuple()
    )
    scan_item.end_time = scan_item.start_time + 10
    scan_item.num_points = 1
    assert (
        str(scan_item)
        == "ScanItem:\n \tStart time: Fri Jun 23 15:11:06 2023\n\tEnd time: Fri Jun 23 15:11:16"
        " 2023\n\tElapsed time: 10.0 s\n\tScan ID: ['scan_id']\n\tScan number: [1]\n\tNumber of"
        " points: 1\n"
    )


def test_scan_item_str_plain():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_item = ScanItem(scan_manager, "queue_id", [1], ["scan_id"], "status")
    assert str(scan_item) == "ScanItem:\n \tScan ID: ['scan_id']\n\tScan number: [1]\n"


def test_emit_data():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_item = ScanItem(scan_manager, "queue_id", [1], ["scan_id"], "status")
    scan_item.bec = mock.Mock()
    scan_item.bec.callbacks = mock.Mock()
    scan_item._run_request_callbacks = mock.Mock()
    msg = messages.ScanMessage(point_id=0, scan_id="scan_id", data={"samx": {"value": 1}})
    scan_item.emit_data(msg)
    scan_item.bec.callbacks.run.assert_called_once_with("scan_segment", msg.content, msg.metadata)
    scan_item._run_request_callbacks.assert_called_once_with(
        "scan_segment", msg.content, msg.metadata
    )


def test_emit_status():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_item = ScanItem(scan_manager, "queue_id", [1], ["scan_id"], "status")
    scan_item.bec = mock.Mock()
    scan_item.bec.callbacks = mock.Mock()
    scan_item._run_request_callbacks = mock.Mock()
    msg = messages.ScanStatusMessage(scan_id="scan_id", status="open", info={"info": "info"})
    scan_item.emit_status(msg)
    scan_item.bec.callbacks.run.assert_called_once_with("scan_status", msg.content, msg.metadata)
    scan_item._run_request_callbacks.assert_called_once_with(
        "scan_status", msg.content, msg.metadata
    )


def test_run_request_callbacks():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_item = ScanItem(scan_manager, "queue_id", [1], ["scan_id"], "status")
    queue_item = QueueItem(
        scan_manager, "queue_id", [{"RID": "requestID"}], "status", None, ["scan_id"]
    )
    with mock.patch("bec_lib.queue_items.update_queue") as mock_update_queue:
        with mock.patch.object(queue_item, "_update_with_buffer") as mock_update_buffer:
            with mock.patch.object(
                scan_manager.queue_storage, "find_queue_item_by_ID"
            ) as mock_find_queue:
                with mock.patch.object(
                    scan_manager.request_storage, "find_request_by_ID"
                ) as mock_find_req:
                    mock_find_queue.return_value = queue_item
                    scan_item._run_request_callbacks("event_type", "data", "metadata")
                    mock_find_req.return_value.callbacks.run.assert_called_once_with(
                        "event_type", "data", "metadata"
                    )


def test_poll_callbacks():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_item = ScanItem(scan_manager, "queue_id", [1], ["scan_id"], "status")
    queue_item = QueueItem(
        scan_manager, "queue_id", [{"RID": "requestID"}], "status", None, ["scan_id"]
    )
    with mock.patch("bec_lib.queue_items.update_queue") as mock_update_queue:
        with mock.patch.object(queue_item, "_update_with_buffer") as mock_update_buffer:
            with mock.patch.object(
                scan_manager.queue_storage, "find_queue_item_by_ID"
            ) as mock_find_queue:
                with mock.patch.object(
                    scan_manager.request_storage, "find_request_by_ID"
                ) as mock_find_req:
                    mock_find_queue.return_value = queue_item
                    scan_item.poll_callbacks()
                    mock_find_req.return_value.callbacks.poll.assert_called_once()


def test_scan_item_eq():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_item1 = ScanItem(scan_manager, "queue_id", [1], ["scan_id"], "status")
    scan_item2 = ScanItem(scan_manager, "queue_id", [1], ["scan_id"], "status")
    assert scan_item1 == scan_item2


def test_scan_item_neq():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_item1 = ScanItem(scan_manager, "queue_id", [1], ["scan_id"], "status")
    scan_item2 = ScanItem(scan_manager, "queue_id", [1], ["scan_id"], "status")
    scan_item2.scan_id = ["scan_id2"]
    assert scan_item1 != scan_item2


def test_update_with_scan_status_aborted():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_manager.scan_storage.update_with_scan_status(
        messages.ScanStatusMessage(scan_id="", status="aborted", info={"info": "info"})
    )


def test_update_with_scan_status_last_scan_number():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_manager.scan_storage.last_scan_number = 0
    with mock.patch.object(scan_manager.scan_storage, "find_scan_by_ID") as mock_find_scan:
        mock_find_scan.return_value = mock.MagicMock()
        scan_manager.scan_storage.update_with_scan_status(
            messages.ScanStatusMessage(scan_id="scan_id", status="aborted", info={"scan_number": 1})
        )
        assert scan_manager.scan_storage.last_scan_number == 1


def test_update_with_scan_status_updates_start_time():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_manager.scan_storage.last_scan_number = 0
    with mock.patch.object(scan_manager.scan_storage, "find_scan_by_ID") as mock_find_scan:
        scan_item = mock.MagicMock()
        mock_find_scan.return_value = scan_item
        scan_manager.scan_storage.update_with_scan_status(
            messages.ScanStatusMessage(
                scan_id="scan_id", status="open", info={"scan_number": 1}, timestamp=10
            )
        )
        assert scan_item.start_time == 10


def test_update_with_scan_status_does_not_update_start_time():
    scan_manager = ScanManager(ConnectorMock(""))
    with mock.patch.object(scan_manager.scan_storage, "find_scan_by_ID") as mock_find_scan:
        scan_item = mock.MagicMock()
        mock_find_scan.return_value = scan_item
        scan_item.start_time = 0
        scan_manager.scan_storage.update_with_scan_status(
            messages.ScanStatusMessage(
                scan_id="scan_id", status="closed", info={"scan_number": 1}, timestamp=10
            )
        )
        assert scan_item.start_time == 0


def test_update_with_scan_status_updates_end_time():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_manager.scan_storage.last_scan_number = 0
    with mock.patch.object(scan_manager.scan_storage, "find_scan_by_ID") as mock_find_scan:
        scan_item = mock.MagicMock()
        mock_find_scan.return_value = scan_item
        scan_manager.scan_storage.update_with_scan_status(
            messages.ScanStatusMessage(
                scan_id="scan_id", status="closed", info={"scan_number": 1}, timestamp=10
            )
        )
        assert scan_item.end_time == 10


def test_update_with_scan_status_does_not_update_end_time():
    scan_manager = ScanManager(ConnectorMock(""))
    with mock.patch.object(scan_manager.scan_storage, "find_scan_by_ID") as mock_find_scan:
        scan_item = mock.MagicMock()
        mock_find_scan.return_value = scan_item
        scan_item.end_time = 0
        scan_manager.scan_storage.update_with_scan_status(
            messages.ScanStatusMessage(
                scan_id="scan_id", status="open", info={"scan_number": 1}, timestamp=10
            )
        )
        assert scan_item.end_time == 0


def test_update_with_scan_status_updates_num_points():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_manager.scan_storage.last_scan_number = 0
    with mock.patch.object(scan_manager.scan_storage, "find_scan_by_ID") as mock_find_scan:
        scan_item = mock.MagicMock()
        mock_find_scan.return_value = scan_item
        scan_manager.scan_storage.update_with_scan_status(
            messages.ScanStatusMessage(
                scan_id="scan_id",
                status="closed",
                info={"scan_number": 1, "num_points": 10},
                timestamp=10,
            )
        )
        assert scan_item.num_points == 10


def test_update_with_scan_status_updates_scan_number():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_manager.scan_storage.last_scan_number = 0
    with mock.patch.object(scan_manager.scan_storage, "find_scan_by_ID") as mock_find_scan:
        scan_item = mock.MagicMock()
        scan_item.scan_number = None
        mock_find_scan.return_value = scan_item
        scan_manager.scan_storage.update_with_scan_status(
            messages.ScanStatusMessage(
                scan_id="scan_id",
                status="closed",
                info={"scan_number": 1, "num_points": 10},
                timestamp=10,
            )
        )
        assert scan_item.scan_number == 1


def test_update_with_scan_status_does_not_update_scan_number():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_manager.scan_storage.last_scan_number = 0
    with mock.patch.object(scan_manager.scan_storage, "find_scan_by_ID") as mock_find_scan:
        scan_item = mock.MagicMock()
        scan_item.scan_number = 2
        mock_find_scan.return_value = scan_item
        scan_manager.scan_storage.update_with_scan_status(
            messages.ScanStatusMessage(
                scan_id="scan_id",
                status="closed",
                info={"scan_number": 1, "num_points": 10},
                timestamp=10,
            )
        )
        assert scan_item.scan_number == 2


def test_update_with_scan_status_adds_scan_def_id():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_manager.scan_storage.last_scan_number = 0
    with mock.patch.object(scan_manager.scan_storage, "find_scan_by_ID") as mock_find_scan:
        scan_item = mock.MagicMock()
        scan_item.open_scan_defs = set()
        mock_find_scan.return_value = scan_item
        scan_manager.scan_storage.update_with_scan_status(
            messages.ScanStatusMessage(
                scan_id="scan_id",
                status="open",
                info={"scan_number": 1, "num_points": 10, "scan_def_id": "scan_def_id"},
                timestamp=10,
            )
        )
        assert "scan_def_id" in scan_item.open_scan_defs


def test_update_with_scan_status_removes_scan_def_id():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_manager.scan_storage.last_scan_number = 0
    with mock.patch.object(scan_manager.scan_storage, "find_scan_by_ID") as mock_find_scan:
        scan_item = mock.MagicMock()
        scan_item.open_scan_defs = set(["scan_def_id"])
        mock_find_scan.return_value = scan_item
        scan_manager.scan_storage.update_with_scan_status(
            messages.ScanStatusMessage(
                scan_id="scan_id",
                status="closed",
                info={"scan_number": 1, "num_points": 10, "scan_def_id": "scan_def_id"},
                timestamp=10,
            )
        )
        assert "scan_def_id" not in scan_item.open_scan_defs


def test_add_scan_segment_emits_data():
    scan_manager = ScanManager(ConnectorMock(""))
    scan_item = mock.MagicMock()
    scan_item.scan_id = "scan_id"
    scan_item.data = ScanData()
    scan_manager.scan_storage.storage.append(scan_item)

    msg = messages.ScanMessage(
        point_id=0, scan_id="scan_id", data={"samx": {"value": 1}}, metadata={"scan_id": "scan_id"}
    )
    scan_manager.scan_storage.add_scan_segment(msg)
    scan_item.emit_data.assert_called_once_with(msg)
    assert scan_item.data.messages == {0: msg}



import subprocess
import sys
from unittest import mock

import IPython
import pytest

from bec_ipython_client import BECIPythonClient, main
from bec_lib import RedisConnector, ServiceConfig


def test_bec_entry_point_globals_and_post_startup(tmpdir):  # , capfd):
    file_to_execute = tmpdir / "post_startup.py"
    with open(file_to_execute, "w") as f:
        f.write(
            """
try:
  completer=get_ipython().Completer
  import sys
  print(completer.all_completions('bec.'), flush=True)
  print(completer.all_completions('BECIP'), flush=True)
finally:
  import os
  import signal
  os.kill(os.getpid(), signal.SIGTERM)
"""
        )
    p = subprocess.Popen(
        [
            sys.executable,
            main.__file__,
            "--nogui",
            "--dont-wait-for-server",
            "--post-startup-file",
            file_to_execute,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )
    output, _ = p.communicate()
    assert "bec.device_manager" in output  # just one of many completions
    assert (
        "BECIPythonClient" not in output
    )  # just to ensure something we don't want is really not there


def test_bec_load_hli_tab_completion(tmpdir):
    """Test that bec hli is loaded and tab completion in the ipython client works"""
    file_to_execute = tmpdir / "post_startup.py"
    with open(file_to_execute, "w") as f:
        f.write(
            """
try:
  completer=get_ipython().Completer
  import sys
  print(completer.all_completions('umv'), flush=True)
  print(completer.all_completions('mv'), flush=True)
finally:
  import os
  import signal
  os.kill(os.getpid(), signal.SIGTERM)
"""
        )
    p = subprocess.Popen(
        [
            sys.executable,
            main.__file__,
            "--nogui",
            "--dont-wait-for-server",
            "--post-startup-file",
            file_to_execute,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )
    output, _ = p.communicate()
    assert "umvr" in output
    assert "umv" in output
    assert "mv" in output
    assert "mvr" in output


def test_ipython_device_completion(bec_client_mock):
    client = bec_client_mock
    # disable history saving (which runs in a separate thread)
    with mock.patch("IPython.core.history.HistoryManager.enabled", False):
        shell = IPython.terminal.interactiveshell.TerminalInteractiveShell.instance()
        shell.user_ns["dev"] = client.device_manager.devices
        completer = IPython.get_ipython().Completer
        assert "dev.samx" in completer.all_completions("dev.sa")
        assert len(completer.all_completions("dev.sa")) == 3


def test_ipython_device_completion_property_access(bec_client_mock):
    client = bec_client_mock
    # disable history saving (which runs in a separate thread)
    with mock.patch("IPython.core.history.HistoryManager.enabled", False):
        shell = IPython.terminal.interactiveshell.TerminalInteractiveShell.instance()
        shell.user_ns["dev"] = client.device_manager.devices
        completer = IPython.get_ipython().Completer
        assert "dev.samx.dummy_controller.some_var" in completer.all_completions(
            "dev.samx.dummy_controller.som"
        )


@pytest.fixture
def service_config():
    return ServiceConfig(
        redis={"host": "localhost", "port": 5000},
        scibec={"host": "localhost", "port": 5001},
        mongodb={"host": "localhost", "port": 50002},
    )


@pytest.fixture
def ipython_client(service_config):
    client = BECIPythonClient(
        config=service_config,
        connector_cls=mock.MagicMock(spec=RedisConnector),
        wait_for_server=False,
    )
    yield client
    client.shutdown()
    client._client._reset_singleton()


def test_bec_ipython_client_start(service_config):
    client = BECIPythonClient(
        config=service_config,
        connector_cls=mock.MagicMock(spec=RedisConnector),
        wait_for_server=True,
    )
    try:
        with mock.patch.object(client._client, "wait_for_service") as wait_for_service:
            with mock.patch.object(client, "_configure_ipython") as configure_ipython:
                with mock.patch.object(client, "_load_scans"):
                    client.start()
                    configure_ipython.assert_called_once()
                    assert mock.call("ScanBundler", mock.ANY) in wait_for_service.call_args_list
                    assert mock.call("ScanServer", mock.ANY) in wait_for_service.call_args_list
                    assert mock.call("DeviceServer", mock.ANY) in wait_for_service.call_args_list
                    assert client.started
    finally:
        client.shutdown()
        client._client._reset_singleton()


def test_bec_update_username_space(ipython_client):
    client = ipython_client
    with mock.patch.object(client, "wait_for_service") as wait_for_service:
        with mock.patch.object(client, "_configure_ipython") as configure_ipython:
            with mock.patch.object(client, "_load_scans"):
                with mock.patch.object(client, "_ip") as mock_ipy:
                    client.start()
                    mock_ipy.user_global_ns = {}
                    my_object = object()
                    client._update_namespace_callback(action="add", ns_objects={"mv": my_object})
                    assert "mv" in mock_ipy.user_global_ns
                    assert mock_ipy.user_global_ns["mv"] == my_object
                    client._update_namespace_callback(action="remove", ns_objects={"mv": my_object})
                    assert "mv" not in mock_ipy.user_global_ns


def test_bec_ipython_client_start_without_bec_services(ipython_client):
    client = ipython_client
    with mock.patch.object(client, "wait_for_service") as wait_for_service:
        with mock.patch.object(client, "_configure_ipython") as configure_ipython:
            with mock.patch.object(client, "_load_scans"):
                client.start()
                configure_ipython.assert_called_once()
                wait_for_service.assert_not_called()



import inspect
import threading
import types
from unittest.mock import Mock, patch

from ophyd import status as ophyd_status_module
from ophyd.status import StatusBase

dummy_thread = Mock(spec=threading.Thread)


class PatchedStatusBase(StatusBase):
    def __init__(self, *args, **kwargs):
        timeout = kwargs.get("timeout", None)
        if not timeout:
            with patch("threading.Thread", dummy_thread):
                super().__init__(*args, **kwargs)
        else:
            super().__init__(*args, **kwargs)

    def set_finished(self, *args, **kwargs):
        super().set_finished(*args, **kwargs)
        if isinstance(self._callback_thread, Mock):
            if self.settle_time > 0:

                def settle_done():
                    self._settled_event.set()
                    self._run_callbacks()

                threading.Timer(self.settle_time, settle_done).start()
            else:
                self._run_callbacks()

    def set_exception(self, *args, **kwargs):
        super().set_exception(*args, **kwargs)
        if isinstance(self._callback_thread, Mock):
            self._run_callbacks()


def monkey_patch_ophyd():
    if ophyd_status_module.StatusBase.__name__ == "PatchedStatusBase":
        # prevent patching multiple times
        return
    for name, klass in inspect.getmembers(
        ophyd_status_module, lambda x: inspect.isclass(x) and StatusBase in x.__mro__
    ):
        mro = klass.mro()
        bases = tuple(PatchedStatusBase if x is StatusBase else x for x in mro)
        new_klass = type("Patched" + name, bases, {})
        setattr(ophyd_status_module, name, new_klass)



"""
This module contains classes for beamline checks, used to check the beamline status.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from bec_lib.device import Device


class BeamlineCondition(ABC):
    """Abstract base class for beamline checks."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.enabled = True

    @property
    @abstractmethod
    def name(self) -> str:
        """Return a name for the beamline check."""

    @abstractmethod
    def run(self) -> bool:
        """Run the beamline check and return True if the beam is okay, False otherwise."""

    @abstractmethod
    def on_failure_msg(self) -> str:
        """Return a message that will be displayed if the beamline check fails."""


class ShutterCondition(BeamlineCondition):
    """Check if the shutter is open."""

    def __init__(self, shutter: Device):
        super().__init__()
        self.shutter = shutter

    @property
    def name(self):
        return "shutter"

    def run(self):
        shutter_val = self.shutter.read(cached=True)
        return shutter_val["value"].lower() == "open"

    def on_failure_msg(self):
        return "Check beam failed: Shutter is closed."


class LightAvailableCondition(BeamlineCondition):
    """Check if the light is available."""

    def __init__(self, machine_status: Device):
        super().__init__()
        self.machine_status = machine_status

    @property
    def name(self):
        return "light_available"

    def run(self):
        machine_status = self.machine_status.read(cached=True)
        return machine_status["value"] in ["Light Available", "Light-Available"]

    def on_failure_msg(self):
        return "Check beam failed: Light not available."


class FastOrbitFeedbackCondition(BeamlineCondition):
    """Check if the fast orbit feedback is running."""

    def __init__(self, sls_fast_orbit_feedback: Device):
        super().__init__()
        self.sls_fast_orbit_feedback = sls_fast_orbit_feedback

    @property
    def name(self):
        return "fast_orbit_feedback"

    def run(self):
        fast_orbit_feedback = self.sls_fast_orbit_feedback.read(cached=True)
        return fast_orbit_feedback["value"] == "running"

    def on_failure_msg(self):
        return "Check beam failed: Fast orbit feedback is not running."



from __future__ import annotations

import time
import uuid
from typing import TYPE_CHECKING

import msgpack

from bec_lib import MessageEndpoints, bec_logger

from .emitter import EmitterBase

logger = bec_logger.logger

if TYPE_CHECKING:
    from .scan_bundler import ScanBundler


class BlueskyEmitter(EmitterBase):
    def __init__(self, scan_bundler: ScanBundler) -> None:
        super().__init__(scan_bundler.connector)
        self.scan_bundler = scan_bundler
        self.bluesky_metadata = {}

    def send_run_start_document(self, scan_id) -> None:
        """Bluesky only: send run start documents."""
        logger.debug(f"sending run start doc for scan_id {scan_id}")
        self.bluesky_metadata[scan_id] = {}
        doc = self._get_run_start_document(scan_id)
        self.bluesky_metadata[scan_id]["start"] = doc
        self.connector.raw_send(MessageEndpoints.bluesky_events(), msgpack.dumps(("start", doc)))
        self.send_descriptor_document(scan_id)

    def _get_run_start_document(self, scan_id) -> dict:
        sb = self.scan_bundler
        doc = {
            "time": time.time(),
            "uid": str(uuid.uuid4()),
            "scan_id": scan_id,
            "queue_id": sb.sync_storage[scan_id]["info"]["queue_id"],
            "scan_id": sb.sync_storage[scan_id]["info"]["scan_number"],
            "motors": tuple(dev.name for dev in sb.scan_motors[scan_id]),
        }
        return doc

    def _get_data_keys(self, scan_id):
        sb = self.scan_bundler
        signals = {}
        for dev in sb.monitored_devices[scan_id]["devices"]:
            # copied from bluesky/callbacks/stream.py:
            signals[dev.name] = sb.device_manager.devices[dev.name]._info.get("describe", {})
        return signals

    def _get_descriptor_document(self, scan_id) -> dict:
        sb = self.scan_bundler
        doc = {
            "run_start": self.bluesky_metadata[scan_id]["start"]["uid"],
            "time": time.time(),
            "data_keys": self._get_data_keys(scan_id),
            "uid": str(uuid.uuid4()),
            "configuration": {},
            "name": "primary",
            "hints": {"samx": {"fields": ["samx"]}, "samy": {"fields": ["samy"]}},
            "object_keys": {
                dev.name: [val["obj_name"] for val in dev._info.get("signals", {})]
                for dev in sb.monitored_devices[scan_id]["devices"]
            },
        }
        return doc

    def send_descriptor_document(self, scan_id) -> None:
        """Bluesky only: send descriptor document"""
        doc = self._get_descriptor_document(scan_id)
        self.bluesky_metadata[scan_id]["descriptor"] = doc
        self.connector.raw_send(
            MessageEndpoints.bluesky_events(), msgpack.dumps(("descriptor", doc))
        )

    def cleanup_storage(self, scan_id):
        """remove old scan_ids to free memory"""

        for storage in ["bluesky_metadata"]:
            try:
                getattr(self, storage).pop(scan_id)
            except KeyError:
                logger.warning(f"Failed to remove {scan_id} from {storage}.")

    def send_bluesky_scan_point(self, scan_id, point_id) -> None:
        self.connector.raw_send(
            MessageEndpoints.bluesky_events(),
            msgpack.dumps(("event", self._prepare_bluesky_event_data(scan_id, point_id))),
        )

    def _prepare_bluesky_event_data(self, scan_id, point_id) -> dict:
        # event = {
        #     "descriptor": "5605e810-bb4e-4e40-b...d45279e3a4",
        #     "time": 1648468217.524021,
        #     "data": {
        #         "det": 1.0,
        #         "motor1": -10.0,
        #         "motor1_setpoint": -10.0,
        #         "motor2": -10.0,
        #         "motor2_setpoint": -10.0,
        #     },
        #     "timestamps": {
        #         "det": 1648468209.868633,
        #         "motor1": 1648468209.862141,
        #         "motor1_setpoint": 1648468209.8607192,
        #         "motor2": 1648468209.864479,
        #         "motor2_setpoint": 1648468209.8629901,
        #     },
        #     "seq_num": 1,
        #     "uid": "ea83a56e-6af2-4b94-9...44dcc36d4e",
        #     "filled": {},
        # }
        sb = self.scan_bundler
        metadata = self.bluesky_metadata[scan_id]
        while not metadata.get("descriptor"):
            time.sleep(0.01)

        bls_event = {
            "descriptor": metadata["descriptor"].get("uid"),
            "time": time.time(),
            "seq_num": point_id,
            "uid": str(uuid.uuid4()),
            "filled": {},
            "data": {},
            "timestamps": {},
        }
        for data_point in sb.sync_storage[scan_id][point_id].values():
            for key, val in data_point.items():
                bls_event["data"][key] = val["value"]
                bls_event["timestamps"][key] = val["timestamp"]
        return bls_event

    def on_cleanup(self, scan_id: str):
        self.cleanup_storage(scan_id)

    def on_init(self, scan_id: str):
        self.send_run_start_document(scan_id)



""" Fixtures for end-to-end tests """

# pylint: skip-file
import os
import pathlib
import platform
import shutil
import tempfile

import pytest
from pytest_redis import factories as pytest_redis_factories
from redis import Redis

from bec_ipython_client import BECIPythonClient
from bec_lib import BECClient, ConfigHelper, RedisConnector, ServiceConfig
from bec_lib.tests.utils import wait_for_empty_queue


@pytest.hookimpl
def pytest_addoption(parser):
    parser.addoption("--start-servers", action="store_true", default=False)
    parser.addoption("--bec-redis-host", action="store", default="localhost")
    parser.addoption("--bec-redis-cmd", action="store", default=None)
    parser.addoption("--flush-redis", action="store_true", default=False)
    parser.addoption("--files-path", action="store", default=None)


redis_server_fixture = None
bec_redis_fixture = None
_start_servers = False
bec_servers_scope = (
    lambda fixture_name, config: config.getoption("--flush-redis") and "function" or "session"
)

services_config_template = """
redis:
  host: %(redis_host)s
  port: %(redis_port)s
mongodb:
  host: "localhost"
  port: 27017
scibec:
  host: http://localhost
  port: 3030
  beamline: TestBeamline
service_config:
  abort_on_ctrl_c: False
  enforce_ACLs: False
  file_writer:
    plugin: default_NeXus_format
    base_path: %(file_writer_base_path)s
  log_writer:
    base_path: %(file_writer_base_path)s
"""


def _check_path(file_path):
    if os.path.exists(file_path):
        return pathlib.Path(file_path)
    else:
        raise RuntimeError(
            f"end2end tests: --files-path directory {repr(file_path)} does not exist"
        )


def _get_tmp_dir():
    # on MacOS, gettempdir() returns path like /var/folders/nj/269977hs0_96bttwj2gs_jhhp48z54/T[...],
    # and if building a Unix socket file (like pytest-redis does to connect to redis) it can
    # exceed the 109 characters limit, so make a special case for MacOS
    return pathlib.Path("/tmp" if platform.system() == "Darwin" else tempfile.gettempdir())


@pytest.hookimpl
def pytest_configure(config):
    global redis_server_fixture
    global bec_redis_fixture
    global _start_servers
    global _bec_servers_scope

    if config.getoption("--start-servers"):
        # configure 'datadir' == where redis Unix socket will go, and .rdb file (if any)
        # try to use specified files path (hope it does not exceed 109 chars) or
        # just use the normal tmp file directory except on MacOS where it must be enforced
        # to /tmp
        user_tmp_path = config.getoption("--files-path")
        if user_tmp_path is not None:
            datadir = _check_path(user_tmp_path)
        else:
            datadir = _get_tmp_dir()
        # session-scoped fixture that starts redis using provided cmd
        redis_server_fixture = pytest_redis_factories.proc.redis_proc(
            executable=config.getoption("--bec-redis-cmd"), datadir=datadir
        )

        if config.getoption("--flush-redis"):
            bec_redis_fixture = pytest_redis_factories.redisdb("redis_server_fixture")
            _bec_servers_scope = "function"  # have to restart servers at each test
        else:
            bec_redis_fixture = redis_server_fixture
    else:
        # do not automatically start redis - bec_redis_fixture will use existing
        # process, will wait for 3 seconds max (must be running already);
        # there is no point checking if we want to flush redis
        # since it would remove available scans which are only populated
        # when scan server starts
        redis_server_fixture = pytest_redis_factories.redis_noproc(
            host=config.getoption("--bec-redis-host"), startup_timeout=3
        )
        bec_redis_fixture = redis_server_fixture

    _start_servers = config.getoption("--start-servers")


@pytest.fixture(scope=bec_servers_scope)
def bec_files_path(request):
    user_tmp_path = request.config.getoption("--files-path")
    if user_tmp_path is not None:
        yield _check_path(user_tmp_path)
    else:
        if request.config.getoption("--flush-redis"):
            request.fixturenames.append("tmp_path")
            yield request.getfixturevalue("tmp_path")
        else:
            request.fixturenames.append("tmp_path_factory")
            yield request.getfixturevalue("tmp_path_factory").mktemp("bec_files")


@pytest.fixture(scope=bec_servers_scope)
def bec_services_config_file_path(bec_files_path):
    return pathlib.Path(bec_files_path / "services_config.yaml")


@pytest.fixture(scope=bec_servers_scope)
def bec_test_config_file_path(bec_files_path):
    return pathlib.Path(bec_files_path / "test_config.yaml")


@pytest.fixture(scope=bec_servers_scope)
def bec_redis_host_port(request, bec_redis_fixture):
    if isinstance(bec_redis_fixture, Redis):
        server_fixture = request.getfixturevalue("redis_server_fixture")
        return server_fixture.host, server_fixture.port
    else:
        return bec_redis_fixture.host, bec_redis_fixture.port


@pytest.fixture(scope=bec_servers_scope)
def bec_servers(
    test_config_yaml_file_path,
    bec_services_config_file_path,
    bec_test_config_file_path,
    bec_files_path,
    bec_redis_host_port,
):
    redis_host, redis_port = bec_redis_host_port
    # ensure configuration files are written where appropriate for tests,
    # i.e. either in /tmp/pytest/... directory, or following user "--files-path"
    # 1) test config (devices...)
    shutil.copyfile(test_config_yaml_file_path, bec_test_config_file_path)
    # 2) path where files are saved
    file_writer_path = bec_services_config_file_path.parent  # / "writer_output"
    # file_writer_path.mkdir(exist_ok=True)
    # 3) services config
    with open(bec_services_config_file_path, "w") as services_config_file:
        services_config_file.write(
            services_config_template
            % {
                "redis_host": redis_host,
                "redis_port": redis_port,
                "file_writer_base_path": file_writer_path,
            }
        )

    if _start_servers:
        from bec_server.bec_server_utils.service_handler import ServiceHandler

        # Start all BEC servers, kill them at the end
        # when no_tmux=True, 'bec_path' indicate the cwd
        # for the process (working directory), i.e. where log files will go
        service_handler = ServiceHandler(
            bec_path=bec_files_path, config_path=bec_services_config_file_path, no_tmux=True
        )
        processes = service_handler.start()
        try:
            yield
        finally:
            for process in processes:
                process.terminate()
            for process in processes:
                os.waitpid(process.pid, 0)
    else:
        # Nothing to do here: servers are supposed to be started externally.
        yield


@pytest.fixture
def bec_ipython_client_with_demo_config(
    bec_redis_fixture, bec_services_config_file_path, bec_servers
):
    config = ServiceConfig(bec_services_config_file_path)
    bec = BECIPythonClient(config, RedisConnector, forced=True)
    bec.start()
    ConfigHelper(bec._client.connector).load_demo_config()
    try:
        yield bec
    finally:
        bec.shutdown()
        bec._client._reset_singleton()


@pytest.fixture
def bec_client_lib_with_demo_config(bec_redis_fixture, bec_services_config_file_path, bec_servers):
    config = ServiceConfig(bec_services_config_file_path)
    bec = BECClient(config, RedisConnector, forced=True, wait_for_server=True)
    bec.start()
    ConfigHelper(bec._client.connector).load_demo_config()
    try:
        yield bec
    finally:
        bec.shutdown()
        bec._client._reset_singleton()


@pytest.fixture
def bec_ipython_client_fixture(bec_ipython_client_with_demo_config):
    bec = bec_ipython_client_with_demo_config
    bec.queue.request_queue_reset()
    bec.queue.request_scan_continuation()
    wait_for_empty_queue(bec)
    yield bec


@pytest.fixture
def bec_client_lib(bec_client_lib_with_demo_config):
    bec = bec_client_lib_with_demo_config
    bec.queue.request_queue_reset()
    bec.queue.request_scan_continuation()
    wait_for_empty_queue(bec)
    yield bec



import textwrap

import numpy as np


class PrettyTable:
    def __init__(self, header: list, padding: int = 18) -> None:
        self.header = header
        self.padding = padding
        self.row_separator = None
        self.format_string = [".2f" for _ in self.header]
        self.format_string[0] = "d"
        self.header_lines = self._get_header_lines()
        self.width = len(self.get_header_separator())
        self.wrapper = textwrap.TextWrapper()
        self.wrapper.width = self.width - 2
        self.wrapper.subsequent_indent = "    "

    def get_header(self) -> str:
        return (
            "|"
            + "|".join(self.aligned_to_center(self.padding, header) for header in self.header)
            + "|"
        )

    def get_row(self, *args) -> str:
        return (
            "|"
            + "|".join(self.aligned_to_center(self.padding, val) for ii, val in enumerate(args))
            + "|"
        )

    def get_row_separator(self) -> str:
        if not self.row_separator:
            self.row_separator = self._get_separator()
        return self.row_separator

    def get_header_separator(self) -> str:
        return self._get_separator(fill="=")

    def _get_separator(self, column_sign="+", fill="-") -> str:
        return "".join(column_sign + fill * self.padding for _ in self.header) + column_sign

    def get_header_lines(self) -> str:
        return self.header_lines

    def _get_header_lines(self) -> str:
        return (
            self.get_row_separator() + "\n" + self.get_header() + "\n" + self.get_header_separator()
        )

    def get_footer(self, footer_text: str):
        footer = self.get_header_separator() + "\n"
        footer_lines = self.wrapper.wrap(" " + footer_text)
        for fl in footer_lines:
            footer += f"{self._get_footer_line(fl)}\n"
        footer += self.get_header_separator() + "\n"
        return footer

    def _get_footer_line(self, footer_line):
        return "| " + footer_line + " " * (self.width - len(footer_line) - 3) + "|"

    @staticmethod
    def aligned_to_center(padding: int, content: str, fill: str = " ") -> str:
        shift = 0

        if content.startswith("-"):
            shift = 1

        return (
            fill * (int(np.ceil((padding - len(content)) / 2)) - shift)
            + content
            + fill * (padding - len(content) - int(np.ceil((padding - len(content)) / 2)) + shift)
        )


if __name__ == "__main__":
    t = PrettyTable(["Acq.", "samx", "samy"])
    t.get_header()
